"""Module which defines the code for the "Models" tab."""

from __future__ import annotations

from typing import TYPE_CHECKING

from collections.abc import Sequence
from functools import partial

import gradio as gr

# NOTE gradio uses pandas for more than typechecking so we need to
# import it here
import pandas as pd  # noqa: TC002

from ultimate_rvc.core.manage.models import (
    delete_all_custom_embedder_models,
    delete_all_custom_pretrained_models,
    delete_all_models,
    delete_all_training_models,
    delete_all_voice_models,
    delete_custom_embedder_models,
    delete_custom_pretrained_models,
    delete_training_models,
    delete_voice_models,
    download_pretrained_model,
    download_voice_model,
    filter_public_models_table,
    get_custom_embedder_model_names,
    get_custom_pretrained_model_names,
    get_pretrained_metadata,
    get_public_model_tags,
    get_training_model_names,
    get_voice_model_names,
    load_public_models_table,
    upload_custom_embedder_model,
    upload_voice_model,
)
from ultimate_rvc.web.common import (
    exception_harness,
    render_msg,
    setup_delete_event,
    update_dropdowns,
)
from ultimate_rvc.web.config.event import ManageModelEventState
from ultimate_rvc.web.tabs.train.multi_step_generation import (
    render as _render_train_multi_step_tab,
)

if TYPE_CHECKING:
    from ultimate_rvc.web.config.main import ModelManagementConfig, TotalConfig


def render(total_config: TotalConfig) -> None:
    """

    Render "Models" tab.

    Parameters
    ----------
    total_config : TotalConfig
        Model containing all component configuration settings for the
        Ultimate RVC web UI.

    """
    tab_config = total_config.management.model
    tab_config.dummy_checkbox.instantiate()
    event_state = ManageModelEventState()

    _render_download_tab(event_state)
    _render_upload_tab(event_state)
    with gr.Tab("Train", elem_id="train-tab"):
        _render_train_multi_step_tab(total_config)
    _render_delete_tab(tab_config, event_state)

    *_, all_model_update = [
        click_event.success(
            partial(update_dropdowns, get_voice_model_names, 5, [], [4]),
            outputs=[
                total_config.song.one_click.voice_model.instance,
                total_config.song.multi_step.voice_model.instance,
                total_config.speech.one_click.voice_model.instance,
                total_config.speech.multi_step.voice_model.instance,
                tab_config.voices.instance,
            ],
            show_progress="hidden",
        )
        for click_event in [
            event_state.download_voice_click.instance,
            event_state.upload_voice_click.instance,
            event_state.delete_voice_click.instance,
            event_state.delete_all_voices_click.instance,
            event_state.delete_all_click.instance,
        ]
    ]

    *_, all_model_update = [
        click_event.success(
            partial(update_dropdowns, get_custom_embedder_model_names, 6, [], [5]),
            outputs=[
                total_config.song.one_click.custom_embedder_model.instance,
                total_config.song.multi_step.custom_embedder_model.instance,
                total_config.speech.one_click.custom_embedder_model.instance,
                total_config.speech.multi_step.custom_embedder_model.instance,
                total_config.training.multi_step.custom_embedder_model.instance,
                tab_config.embedders.instance,
            ],
            show_progress="hidden",
        )
        for click_event in [
            event_state.upload_embedder_click.instance,
            event_state.delete_embedder_click.instance,
            event_state.delete_all_embedders_click.instance,
            all_model_update,
        ]
    ]

    *_, all_model_update = [
        click_event.success(
            partial(update_dropdowns, get_custom_pretrained_model_names, 2, [], [1]),
            outputs=[
                total_config.training.multi_step.custom_pretrained_model.instance,
                tab_config.pretraineds.instance,
            ],
            show_progress="hidden",
        )
        for click_event in [
            event_state.download_pretrained_click.instance,
            event_state.delete_pretrained_click.instance,
            event_state.delete_all_pretraineds_click.instance,
            all_model_update,
        ]
    ]

    for click_event in [
        event_state.delete_trained_click.instance,
        event_state.delete_all_trained_click.instance,
        all_model_update,
    ]:
        click_event.success(
            partial(update_dropdowns, get_training_model_names, 4, [], [0, 3]),
            outputs=[
                total_config.training.multi_step.preprocess_model.instance,
                total_config.training.multi_step.extract_model.instance,
                total_config.training.multi_step.train_model.instance,
                tab_config.traineds.instance,
            ],
            show_progress="hidden",
        )


def _render_download_tab(event_state: ManageModelEventState) -> None:
    with gr.Tab("Download"):
        with gr.Accordion("Voice models"):
            with gr.Accordion("View public models table", open=False):
                with gr.Accordion("HOW TO USE", open=False):
                    gr.Markdown("")
                    gr.Markdown(
                        "- Filter voice models by selecting one or more tags and/or"
                        " providing a search query.",
                    )
                    gr.Markdown(
                        "- Select a row in the table to autofill the name and URL for"
                        " the given voice model in the form fields below.",
                    )
                with gr.Row():
                    search_query = gr.Textbox(label="Search query")
                    tags = gr.CheckboxGroup(
                        label="Tags",
                        value=[],
                        choices=get_public_model_tags(),
                    )
                with gr.Row():
                    public_models_table = gr.Dataframe(
                        label="Public models table",
                        value=load_public_models_table([]),
                        headers=[
                            "Name",
                            "Description",
                            "Tags",
                            "Credit",
                            "Added",
                            "URL",
                        ],
                        interactive=False,
                    )
                # We are updating the table here instead of doing it
                # implicitly using value=_filter_public_models_table and
                # inputs=[tags, search_query] when instantiating
                # gr.Dataframe because that does not work with reload
                # mode due to a bug.
                gr.on(
                    triggers=[search_query.change, tags.change],
                    fn=_filter_public_models_table,
                    inputs=[tags, search_query],
                    outputs=public_models_table,
                )

            with gr.Row():
                voice_model_url = gr.Textbox(
                    label="Model URL",
                    info=(
                        "Should point to a zip file containing a .pth model file and"
                        " optionally also an .index file."
                    ),
                )
                voice_model_name = gr.Textbox(
                    label="Model name",
                    info="Enter a unique name for the voice model.",
                )

            with gr.Row(equal_height=True):
                download_voice_btn = gr.Button(
                    "Download 🌐",
                    variant="primary",
                    scale=19,
                )
                download_voice_msg = gr.Textbox(
                    label="Output message",
                    interactive=False,
                    scale=20,
                )

            public_models_table.select(
                _autofill_model_name_and_url,
                inputs=public_models_table,
                outputs=[voice_model_name, voice_model_url],
                show_progress="hidden",
            )

            event_state.download_voice_click.instance = download_voice_btn.click(
                exception_harness(download_voice_model),
                inputs=[voice_model_url, voice_model_name],
                outputs=download_voice_msg,
            ).success(
                partial(render_msg, "[+] Succesfully downloaded voice model!"),
                outputs=download_voice_msg,
                show_progress="hidden",
            )
        with gr.Accordion("Pretrained models", open=False):
            with gr.Row():
                pretrained_model_metadata = get_pretrained_metadata()
                pretrained_model = gr.Dropdown(
                    label="Pretrained model",
                    info="Select the pretrained model you want to download.",
                    value=pretrained_model_metadata.default_name,
                    choices=pretrained_model_metadata.names,
                )
                pretrained_sample_rate = gr.Dropdown(
                    label="Sample rate",
                    info="Select the sample rate for the pretrained model.",
                    value=pretrained_model_metadata.default_sample_rate,
                    choices=pretrained_model_metadata.default_sample_rates,
                )

                pretrained_model.change(
                    _update_pretrained_sample_rates,
                    inputs=pretrained_model,
                    outputs=pretrained_sample_rate,
                    show_progress="hidden",
                )
            with gr.Row(equal_height=True):
                download_pretrained_btn = gr.Button(
                    "Download 🌐",
                    variant="primary",
                    scale=19,
                )
                download_pretrained_msg = gr.Textbox(
                    label="Output message",
                    interactive=False,
                    scale=20,
                )
            event_state.download_pretrained_click.instance = (
                download_pretrained_btn.click(
                    exception_harness(download_pretrained_model),
                    inputs=[pretrained_model, pretrained_sample_rate],
                    outputs=download_pretrained_msg,
                ).success(
                    partial(render_msg, "[+] Succesfully downloaded pretrained model!"),
                    outputs=download_pretrained_msg,
                    show_progress="hidden",
                )
            )


def _render_upload_tab(event_state: ManageModelEventState) -> None:
    with gr.Tab("Upload"):
        with gr.Accordion("Voice models", open=True):
            with gr.Accordion("HOW TO USE", open=False):
                gr.Markdown("")
                gr.Markdown(
                    "1. Find the .pth file for a locally trained RVC model (e.g. in"
                    " your local weights folder) and optionally also a corresponding"
                    " .index file (e.g. in your logs/[name] folder)",
                )
                gr.Markdown(
                    "2. Upload the files directly or save them to a folder, then"
                    " compress that folder and upload the resulting .zip file",
                )
                gr.Markdown("3. Enter a unique name for the uploaded model")
                gr.Markdown("4. Click 'Upload'")

            with gr.Row():
                voice_model_files = gr.File(
                    label="Files",
                    file_count="multiple",
                    file_types=[".zip", ".pth", ".index"],
                )

                local_voice_model_name = gr.Textbox(label="Model name")

            with gr.Row(equal_height=True):
                upload_voice_btn = gr.Button("Upload", variant="primary", scale=19)
                upload_voice_msg = gr.Textbox(
                    label="Output message",
                    interactive=False,
                    scale=20,
                )
                event_state.upload_voice_click.instance = upload_voice_btn.click(
                    exception_harness(upload_voice_model),
                    inputs=[voice_model_files, local_voice_model_name],
                    outputs=upload_voice_msg,
                ).success(
                    partial(render_msg, "[+] Successfully uploaded voice model!"),
                    outputs=upload_voice_msg,
                    show_progress="hidden",
                )
        with gr.Accordion("Custom embedder models", open=False):
            with gr.Accordion("HOW TO USE", open=False):
                gr.Markdown("")
                gr.Markdown(
                    "1. Find the config.json file and pytorch_model.bin file for a"
                    " custom embedder model stored locally.",
                )
                gr.Markdown(
                    "2. Upload the files directly or save them to a folder, then"
                    " compress that folder and upload the resulting .zip file",
                )
                gr.Markdown("3. Enter a unique name for the uploaded embedder model")
                gr.Markdown("4. Click 'Upload'")

            with gr.Row():
                embedder_files = gr.File(
                    label="Files",
                    file_count="multiple",
                    file_types=[".zip", ".json", ".bin"],
                )

                local_embedder_name = gr.Textbox(label="Model name")

            with gr.Row(equal_height=True):
                upload_embedder_btn = gr.Button("Upload", variant="primary", scale=19)
                upload_embedder_msg = gr.Textbox(
                    label="Output message",
                    interactive=False,
                    scale=20,
                )
                event_state.upload_embedder_click.instance = upload_embedder_btn.click(
                    exception_harness(upload_custom_embedder_model),
                    inputs=[embedder_files, local_embedder_name],
                    outputs=upload_embedder_msg,
                ).success(
                    partial(
                        render_msg,
                        "[+] Successfully uploaded custom embedder model!",
                    ),
                    outputs=upload_embedder_msg,
                    show_progress="hidden",
                )


def _render_delete_tab(
    tab_config: ModelManagementConfig,
    event_state: ManageModelEventState,
) -> None:
    with gr.Tab("Delete"):
        _render_voices_accordion(tab_config, event_state)
        _render_embedders_accordion(tab_config, event_state)
        _render_pretraineds_accordion(tab_config, event_state)
        _render_traineds_accordion(tab_config, event_state)
        _render_all_accordion(tab_config, event_state)


def _render_voices_accordion(
    tab_config: ModelManagementConfig,
    event_state: ManageModelEventState,
) -> None:
    with gr.Accordion("Voice models", open=False), gr.Row():
        with gr.Column():
            tab_config.voices.instance.render()
            delete_voice_btn = gr.Button("Delete selected", variant="secondary")
            delete_all_voice_btn = gr.Button("Delete all", variant="primary")
        with gr.Column():
            delete_voice_msg = gr.Textbox(label="Output message", interactive=False)

    event_state.delete_voice_click.instance = setup_delete_event(
        delete_voice_btn,
        delete_voice_models,
        [tab_config.dummy_checkbox.instance, tab_config.voices.instance],
        delete_voice_msg,
        "Are you sure you want to delete the selected voice models?",
        "[-] Successfully deleted selected voice models!",
    )

    event_state.delete_all_voices_click.instance = setup_delete_event(
        delete_all_voice_btn,
        delete_all_voice_models,
        [tab_config.dummy_checkbox.instance],
        delete_voice_msg,
        "Are you sure you want to delete all voice models?",
        "[-] Successfully deleted all voice models!",
    )


def _render_embedders_accordion(
    tab_config: ModelManagementConfig,
    event_state: ManageModelEventState,
) -> None:
    with gr.Accordion("Custom embedder models", open=False), gr.Row():
        with gr.Column():
            tab_config.embedders.instance.render()
            delete_embedder_btn = gr.Button("Delete selected", variant="secondary")
            delete_all_embedder_btn = gr.Button("Delete all", variant="primary")
        with gr.Column():
            delete_embedder_msg = gr.Textbox(label="Output message", interactive=False)

    event_state.delete_embedder_click.instance = setup_delete_event(
        delete_embedder_btn,
        delete_custom_embedder_models,
        [tab_config.dummy_checkbox.instance, tab_config.embedders.instance],
        delete_embedder_msg,
        "Are you sure you want to delete the selected custom embedder models?",
        "[-] Successfully deleted selected custom embedder models!",
    )

    event_state.delete_all_embedders_click.instance = setup_delete_event(
        delete_all_embedder_btn,
        delete_all_custom_embedder_models,
        [tab_config.dummy_checkbox.instance],
        delete_embedder_msg,
        "Are you sure you want to delete all custom embedder models?",
        "[-] Successfully deleted all custom embedder models!",
    )


def _render_pretraineds_accordion(
    tab_config: ModelManagementConfig,
    event_state: ManageModelEventState,
) -> None:
    with gr.Accordion("Custom pretrained models", open=False), gr.Row():
        with gr.Column():
            tab_config.pretraineds.instance.render()
            delete_pretrained_btn = gr.Button("Delete selected", variant="secondary")
            delete_all_pretrained_btn = gr.Button("Delete all", variant="primary")
        with gr.Column():
            delete_pretrained_msg = gr.Textbox(
                label="Output message",
                interactive=False,
            )

    event_state.delete_pretrained_click.instance = setup_delete_event(
        delete_pretrained_btn,
        delete_custom_pretrained_models,
        [tab_config.dummy_checkbox.instance, tab_config.pretraineds.instance],
        delete_pretrained_msg,
        "Are you sure you want to delete the selected custom pretrained models?",
        "[-] Successfully deleted selected custom pretrained models!",
    )
    event_state.delete_all_pretraineds_click.instance = setup_delete_event(
        delete_all_pretrained_btn,
        delete_all_custom_pretrained_models,
        [tab_config.dummy_checkbox.instance],
        delete_pretrained_msg,
        "Are you sure you want to delete all custom pretrained models?",
        "[-] Successfully deleted all custom pretrained models!",
    )


def _render_traineds_accordion(
    tab_config: ModelManagementConfig,
    event_state: ManageModelEventState,
) -> None:
    with gr.Accordion("Training models", open=False), gr.Row():
        with gr.Column():
            tab_config.traineds.instance.render()
            delete_train_btn = gr.Button("Delete selected", variant="secondary")
            delete_all_train_btn = gr.Button("Delete all", variant="primary")
        with gr.Column():
            delete_train_msg = gr.Textbox(label="Output message", interactive=False)

    event_state.delete_trained_click.instance = setup_delete_event(
        delete_train_btn,
        delete_training_models,
        [tab_config.dummy_checkbox.instance, tab_config.traineds.instance],
        delete_train_msg,
        "Are you sure you want to delete the selected training models?",
        "[-] Successfully deleted selected training models!",
    )

    event_state.delete_all_trained_click.instance = setup_delete_event(
        delete_all_train_btn,
        delete_all_training_models,
        [tab_config.dummy_checkbox.instance],
        delete_train_msg,
        "Are you sure you want to delete all training models?",
        "[-] Successfully deleted all training models!",
    )


def _render_all_accordion(
    tab_config: ModelManagementConfig,
    event_state: ManageModelEventState,
) -> None:
    with gr.Accordion("All models"), gr.Row(equal_height=True):
        delete_all_btn = gr.Button("Delete", variant="primary")
        delete_all_msg = gr.Textbox(label="Output message", interactive=False)

    event_state.delete_all_click.instance = setup_delete_event(
        delete_all_btn,
        delete_all_models,
        [tab_config.dummy_checkbox.instance],
        delete_all_msg,
        "Are you sure you want to delete all models?",
        "[-] Successfully deleted all models!",
    )


def _filter_public_models_table(tags: Sequence[str], query: str) -> gr.Dataframe:
    """
    Filter table containing metadata of public voice models by tags and
    a search query.

    Parameters
    ----------
    tags : Sequence[str]
        Tags to filter the metadata table by.
    query : str
        Search query to filter the metadata table by.

    Returns
    -------
    gr.Dataframe
        The filtered table rendered in a Gradio dataframe.

    """
    models_table = filter_public_models_table(tags, query)
    return gr.Dataframe(value=models_table)


def _autofill_model_name_and_url(
    public_models_table: pd.DataFrame,
    select_event: gr.SelectData,
) -> tuple[gr.Textbox, gr.Textbox]:
    """
    Autofill two textboxes with respectively the name and URL that is
    saved in the currently selected row of the public models table.

    Parameters
    ----------
    public_models_table : pd.DataFrame
        The public models table saved in a Pandas dataframe.
    select_event : gr.SelectData
        Event containing the index of the currently selected row in the
        public models table.

    Returns
    -------
    name : gr.Textbox
        The textbox containing the model name.

    url : gr.Textbox
        The textbox containing the model URL.

    Raises
    ------
    TypeError
        If the index in the provided event is not a sequence.

    """
    event_index: int | Sequence[int] = select_event.index
    if not isinstance(event_index, Sequence):
        err_msg = (
            f"Expected a sequence of indices but got {type(event_index)} from the"
            " provided event."
        )
        raise TypeError(err_msg)
    event_index = event_index[0]
    url = public_models_table.loc[event_index, "URL"]
    name = public_models_table.loc[event_index, "Name"]
    if isinstance(url, str) and isinstance(name, str):
        return gr.Textbox(value=name), gr.Textbox(value=url)
    err_msg = (
        "Expected model name and URL to be strings but got"
        f" {type(name)} and {type(url)} respectively."
    )
    raise TypeError(err_msg)


def _update_pretrained_sample_rates(name: str) -> gr.Dropdown:
    """
    Update the dropdown for pretrained sample rates based on the
    selected pretrained model.

    Parameters
    ----------
    name : str
        The name of the selected pretrained model.

    Returns
    -------
    pretrained_sample_rate : gr.Dropdown
        The updated dropdown for pretrained sample rates.

    """
    pretrained_metadata = get_pretrained_metadata()
    default, sample_rates = pretrained_metadata.get_sample_rates_with_default(name)
    return gr.Dropdown(value=default, choices=sample_rates)



import math

import torch

from ultimate_rvc.rvc.lib.algorithm.commons import convert_pad_shape


class MultiHeadAttention(torch.nn.Module):
    """
    Multi-head attention module with optional relative positional encoding and proximal bias.

    Args:
        channels (int): Number of input channels.
        out_channels (int): Number of output channels.
        n_heads (int): Number of attention heads.
        p_dropout (float, optional): Dropout probability. Defaults to 0.0.
        window_size (int, optional): Window size for relative positional encoding. Defaults to None.
        heads_share (bool, optional): Whether to share relative positional embeddings across heads. Defaults to True.
        block_length (int, optional): Block length for local attention. Defaults to None.
        proximal_bias (bool, optional): Whether to use proximal bias in self-attention. Defaults to False.
        proximal_init (bool, optional): Whether to initialize the key projection weights the same as query projection weights. Defaults to False.

    """

    def __init__(
        self,
        channels: int,
        out_channels: int,
        n_heads: int,
        p_dropout: float = 0.0,
        window_size: int = None,
        heads_share: bool = True,
        block_length: int = None,
        proximal_bias: bool = False,
        proximal_init: bool = False,
    ):
        super().__init__()
        assert (
            channels % n_heads == 0
        ), "Channels must be divisible by the number of heads."

        self.channels = channels
        self.out_channels = out_channels
        self.n_heads = n_heads
        self.k_channels = channels // n_heads
        self.window_size = window_size
        self.block_length = block_length
        self.proximal_bias = proximal_bias

        # Define projections
        self.conv_q = torch.nn.Conv1d(channels, channels, 1)
        self.conv_k = torch.nn.Conv1d(channels, channels, 1)
        self.conv_v = torch.nn.Conv1d(channels, channels, 1)
        self.conv_o = torch.nn.Conv1d(channels, out_channels, 1)

        self.drop = torch.nn.Dropout(p_dropout)

        # Relative positional encodings
        if window_size:
            n_heads_rel = 1 if heads_share else n_heads
            rel_stddev = self.k_channels**-0.5
            self.emb_rel_k = torch.nn.Parameter(
                torch.randn(n_heads_rel, 2 * window_size + 1, self.k_channels)
                * rel_stddev,
            )
            self.emb_rel_v = torch.nn.Parameter(
                torch.randn(n_heads_rel, 2 * window_size + 1, self.k_channels)
                * rel_stddev,
            )

        # Initialize weights
        torch.nn.init.xavier_uniform_(self.conv_q.weight)
        torch.nn.init.xavier_uniform_(self.conv_k.weight)
        torch.nn.init.xavier_uniform_(self.conv_v.weight)
        torch.nn.init.xavier_uniform_(self.conv_o.weight)

        if proximal_init:
            with torch.no_grad():
                self.conv_k.weight.copy_(self.conv_q.weight)
                self.conv_k.bias.copy_(self.conv_q.bias)

    def forward(self, x, c, attn_mask=None):
        # Compute query, key, value projections
        q, k, v = self.conv_q(x), self.conv_k(c), self.conv_v(c)

        # Compute attention
        x, self.attn = self.attention(q, k, v, mask=attn_mask)

        # Final output projection
        return self.conv_o(x)

    def attention(self, query, key, value, mask=None):
        # Reshape and compute scaled dot-product attention
        b, d, t_s, t_t = (*key.size(), query.size(2))
        query = query.view(b, self.n_heads, self.k_channels, t_t).transpose(2, 3)
        key = key.view(b, self.n_heads, self.k_channels, t_s).transpose(2, 3)
        value = value.view(b, self.n_heads, self.k_channels, t_s).transpose(2, 3)

        scores = torch.matmul(query / math.sqrt(self.k_channels), key.transpose(-2, -1))

        if self.window_size:
            assert t_s == t_t, "Relative attention only supports self-attention."
            scores += self._compute_relative_scores(query, t_s)

        if self.proximal_bias:
            assert t_s == t_t, "Proximal bias only supports self-attention."
            scores += self._attention_bias_proximal(t_s).to(scores.device, scores.dtype)

        if mask is not None:
            scores = scores.masked_fill(mask == 0, -1e4)
            if self.block_length:
                block_mask = (
                    torch.ones_like(scores)
                    .triu(-self.block_length)
                    .tril(self.block_length)
                )
                scores = scores.masked_fill(block_mask == 0, -1e4)

        # Apply softmax and dropout
        p_attn = self.drop(torch.nn.functional.softmax(scores, dim=-1))

        # Compute attention output
        output = torch.matmul(p_attn, value)

        if self.window_size:
            output += self._apply_relative_values(p_attn, t_s)

        return output.transpose(2, 3).contiguous().view(b, d, t_t), p_attn

    def _compute_relative_scores(self, query, length):
        rel_emb = self._get_relative_embeddings(self.emb_rel_k, length)
        rel_logits = self._matmul_with_relative_keys(
            query / math.sqrt(self.k_channels),
            rel_emb,
        )
        return self._relative_position_to_absolute_position(rel_logits)

    def _apply_relative_values(self, p_attn, length):
        rel_weights = self._absolute_position_to_relative_position(p_attn)
        rel_emb = self._get_relative_embeddings(self.emb_rel_v, length)
        return self._matmul_with_relative_values(rel_weights, rel_emb)

    # Helper methods
    def _matmul_with_relative_values(self, x, y):
        return torch.matmul(x, y.unsqueeze(0))

    def _matmul_with_relative_keys(self, x, y):
        return torch.matmul(x, y.unsqueeze(0).transpose(-2, -1))

    def _get_relative_embeddings(self, embeddings, length):
        pad_length = max(length - (self.window_size + 1), 0)
        start = max((self.window_size + 1) - length, 0)
        end = start + 2 * length - 1

        if pad_length > 0:
            embeddings = torch.nn.functional.pad(
                embeddings,
                convert_pad_shape([[0, 0], [pad_length, pad_length], [0, 0]]),
            )
        return embeddings[:, start:end]

    def _relative_position_to_absolute_position(self, x):
        batch, heads, length, _ = x.size()
        x = torch.nn.functional.pad(
            x,
            convert_pad_shape([[0, 0], [0, 0], [0, 0], [0, 1]]),
        )
        x_flat = x.view(batch, heads, length * 2 * length)
        x_flat = torch.nn.functional.pad(
            x_flat,
            convert_pad_shape([[0, 0], [0, 0], [0, length - 1]]),
        )
        return x_flat.view(batch, heads, length + 1, 2 * length - 1)[
            :,
            :,
            :length,
            length - 1 :,
        ]

    def _absolute_position_to_relative_position(self, x):
        batch, heads, length, _ = x.size()
        x = torch.nn.functional.pad(
            x,
            convert_pad_shape([[0, 0], [0, 0], [0, 0], [0, length - 1]]),
        )
        x_flat = x.view(batch, heads, length**2 + length * (length - 1))
        x_flat = torch.nn.functional.pad(
            x_flat,
            convert_pad_shape([[0, 0], [0, 0], [length, 0]]),
        )
        return x_flat.view(batch, heads, length, 2 * length)[:, :, :, 1:]

    def _attention_bias_proximal(self, length):
        r = torch.arange(length, dtype=torch.float32)
        diff = r.unsqueeze(0) - r.unsqueeze(1)
        return -torch.log1p(torch.abs(diff)).unsqueeze(0).unsqueeze(0)


class FFN(torch.nn.Module):
    """
    Feed-forward network module.

    Args:
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels.
        filter_channels (int): Number of filter channels in the convolution layers.
        kernel_size (int): Kernel size of the convolution layers.
        p_dropout (float, optional): Dropout probability. Defaults to 0.0.
        activation (str, optional): Activation function to use. Defaults to None.
        causal (bool, optional): Whether to use causal padding in the convolution layers. Defaults to False.

    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        filter_channels: int,
        kernel_size: int,
        p_dropout: float = 0.0,
        activation: str = None,
        causal: bool = False,
    ):
        super().__init__()
        self.padding_fn = self._causal_padding if causal else self._same_padding

        self.conv_1 = torch.nn.Conv1d(in_channels, filter_channels, kernel_size)
        self.conv_2 = torch.nn.Conv1d(filter_channels, out_channels, kernel_size)
        self.drop = torch.nn.Dropout(p_dropout)

        self.activation = activation

    def forward(self, x, x_mask):
        x = self.conv_1(self.padding_fn(x * x_mask))
        x = self._apply_activation(x)
        x = self.drop(x)
        x = self.conv_2(self.padding_fn(x * x_mask))
        return x * x_mask

    def _apply_activation(self, x):
        if self.activation == "gelu":
            return x * torch.sigmoid(1.702 * x)
        return torch.relu(x)

    def _causal_padding(self, x):
        pad_l, pad_r = self.conv_1.kernel_size[0] - 1, 0
        return torch.nn.functional.pad(
            x,
            convert_pad_shape([[0, 0], [0, 0], [pad_l, pad_r]]),
        )

    def _same_padding(self, x):
        pad = (self.conv_1.kernel_size[0] - 1) // 2
        return torch.nn.functional.pad(
            x,
            convert_pad_shape([[0, 0], [0, 0], [pad, pad]]),
        )



from typing import TYPE_CHECKING

import gc
import os
import re
import sys

import numpy as np
from scipy import signal

import faiss
import torch
import torch.nn.functional as F
from torch import Tensor

import librosa

now_dir = os.getcwd()
sys.path.append(now_dir)

import lazy_loader as lazy

import logging

from ultimate_rvc.common import RVC_MODELS_DIR
from ultimate_rvc.rvc.lib.predictors.FCPE import FCPEF0Predictor
from ultimate_rvc.rvc.lib.predictors.RMVPE import RMVPE0Predictor

if TYPE_CHECKING:
    import torchcrepe
else:
    torchcrepe = lazy.load("torchcrepe")

# logging.getLogger("faiss").setLevel(logging.WARNING)
logger = logging.getLogger(__name__)

# Constants for high-pass filter
FILTER_ORDER = 5
CUTOFF_FREQUENCY = 48  # Hz
SAMPLE_RATE = 16000  # Hz
bh, ah = signal.butter(
    N=FILTER_ORDER,
    Wn=CUTOFF_FREQUENCY,
    btype="high",
    fs=SAMPLE_RATE,
)

input_audio_path2wav = {}


class AudioProcessor:
    """
    A class for processing audio signals, specifically for adjusting RMS levels.
    """

    def change_rms(
        source_audio: np.ndarray,
        source_rate: int,
        target_audio: np.ndarray,
        target_rate: int,
        rate: float,
    ) -> np.ndarray:
        """
        Adjust the RMS level of target_audio to match the RMS of source_audio, with a given blending rate.

        Args:
            source_audio: The source audio signal as a NumPy array.
            source_rate: The sampling rate of the source audio.
            target_audio: The target audio signal to adjust.
            target_rate: The sampling rate of the target audio.
            rate: The blending rate between the source and target RMS levels.

        """
        # Calculate RMS of both audio data
        rms1 = librosa.feature.rms(
            y=source_audio,
            frame_length=source_rate // 2 * 2,
            hop_length=source_rate // 2,
        )
        rms2 = librosa.feature.rms(
            y=target_audio,
            frame_length=target_rate // 2 * 2,
            hop_length=target_rate // 2,
        )

        # Interpolate RMS to match target audio length
        rms1 = F.interpolate(
            torch.from_numpy(rms1).float().unsqueeze(0),
            size=target_audio.shape[0],
            mode="linear",
        ).squeeze()
        rms2 = F.interpolate(
            torch.from_numpy(rms2).float().unsqueeze(0),
            size=target_audio.shape[0],
            mode="linear",
        ).squeeze()
        rms2 = torch.maximum(rms2, torch.zeros_like(rms2) + 1e-6)

        # Adjust target audio RMS based on the source audio RMS
        adjusted_audio = (
            target_audio
            * (torch.pow(rms1, 1 - rate) * torch.pow(rms2, rate - 1)).numpy()
        )
        return adjusted_audio


class Autotune:
    """
    A class for applying autotune to a given fundamental frequency (F0) contour.
    """

    def __init__(self, ref_freqs):
        """
        Initializes the Autotune class with a set of reference frequencies.

        Args:
            ref_freqs: A list of reference frequencies representing musical notes.

        """
        self.ref_freqs = ref_freqs
        self.note_dict = self.ref_freqs  # No interpolation needed

    def autotune_f0(self, f0, f0_autotune_strength):
        """
        Autotunes a given F0 contour by snapping each frequency to the closest reference frequency.

        Args:
            f0: The input F0 contour as a NumPy array.

        """
        autotuned_f0 = np.zeros_like(f0)
        for i, freq in enumerate(f0):
            closest_note = min(self.note_dict, key=lambda x: abs(x - freq))
            autotuned_f0[i] = freq + (closest_note - freq) * f0_autotune_strength
        return autotuned_f0


class Pipeline:
    """
    The main pipeline class for performing voice conversion, including preprocessing, F0 estimation,
    voice conversion using a model, and post-processing.
    """

    def __init__(self, tgt_sr, config):
        """
        Initializes the Pipeline class with target sampling rate and configuration parameters.

        Args:
            tgt_sr: The target sampling rate for the output audio.
            config: A configuration object containing various parameters for the pipeline.

        """
        self.x_pad = config.x_pad
        self.x_query = config.x_query
        self.x_center = config.x_center
        self.x_max = config.x_max
        self.sample_rate = 16000
        self.window = 160
        self.t_pad = self.sample_rate * self.x_pad
        self.t_pad_tgt = tgt_sr * self.x_pad
        self.t_pad2 = self.t_pad * 2
        self.t_query = self.sample_rate * self.x_query
        self.t_center = self.sample_rate * self.x_center
        self.t_max = self.sample_rate * self.x_max
        self.time_step = self.window / self.sample_rate * 1000
        self.f0_min = 50
        self.f0_max = 1100
        self.f0_mel_min = 1127 * np.log(1 + self.f0_min / 700)
        self.f0_mel_max = 1127 * np.log(1 + self.f0_max / 700)
        self.device = config.device
        self.ref_freqs = [
            49.00,  # G1
            51.91,  # G#1 / Ab1
            55.00,  # A1
            58.27,  # A#1 / Bb1
            61.74,  # B1
            65.41,  # C2
            69.30,  # C#2 / Db2
            73.42,  # D2
            77.78,  # D#2 / Eb2
            82.41,  # E2
            87.31,  # F2
            92.50,  # F#2 / Gb2
            98.00,  # G2
            103.83,  # G#2 / Ab2
            110.00,  # A2
            116.54,  # A#2 / Bb2
            123.47,  # B2
            130.81,  # C3
            138.59,  # C#3 / Db3
            146.83,  # D3
            155.56,  # D#3 / Eb3
            164.81,  # E3
            174.61,  # F3
            185.00,  # F#3 / Gb3
            196.00,  # G3
            207.65,  # G#3 / Ab3
            220.00,  # A3
            233.08,  # A#3 / Bb3
            246.94,  # B3
            261.63,  # C4
            277.18,  # C#4 / Db4
            293.66,  # D4
            311.13,  # D#4 / Eb4
            329.63,  # E4
            349.23,  # F4
            369.99,  # F#4 / Gb4
            392.00,  # G4
            415.30,  # G#4 / Ab4
            440.00,  # A4
            466.16,  # A#4 / Bb4
            493.88,  # B4
            523.25,  # C5
            554.37,  # C#5 / Db5
            587.33,  # D5
            622.25,  # D#5 / Eb5
            659.25,  # E5
            698.46,  # F5
            739.99,  # F#5 / Gb5
            783.99,  # G5
            830.61,  # G#5 / Ab5
            880.00,  # A5
            932.33,  # A#5 / Bb5
            987.77,  # B5
            1046.50,  # C6
        ]
        self.autotune = Autotune(self.ref_freqs)
        self.note_dict = self.autotune.note_dict
        self.model_rmvpe = RMVPE0Predictor(
            os.path.join(str(RVC_MODELS_DIR), "predictors", "rmvpe.pt"),
            device=self.device,
        )

    def get_f0_crepe(
        self,
        x,
        f0_min,
        f0_max,
        p_len,
        hop_length,
        model="full",
    ):
        """
        Estimates the fundamental frequency (F0) of a given audio signal using the Crepe model.

        Args:
            x: The input audio signal as a NumPy array.
            f0_min: Minimum F0 value to consider.
            f0_max: Maximum F0 value to consider.
            p_len: Desired length of the F0 output.
            hop_length: Hop length for the Crepe model.
            model: Crepe model size to use ("full" or "tiny").

        """
        x = x.astype(np.float32)
        x /= np.quantile(np.abs(x), 0.999)
        audio = torch.from_numpy(x).to(self.device, copy=True)
        audio = torch.unsqueeze(audio, dim=0)
        if audio.ndim == 2 and audio.shape[0] > 1:
            audio = torch.mean(audio, dim=0, keepdim=True).detach()
        audio = audio.detach()
        pitch: Tensor = torchcrepe.predict(
            audio,
            self.sample_rate,
            hop_length,
            f0_min,
            f0_max,
            model,
            batch_size=hop_length * 2,
            device=self.device,
            pad=True,
        )
        p_len = p_len or x.shape[0] // hop_length
        source = np.array(pitch.squeeze(0).cpu().float().numpy())
        source[source < 0.001] = np.nan
        target = np.interp(
            np.arange(0, len(source) * p_len, len(source)) / p_len,
            np.arange(0, len(source)),
            source,
        )
        f0 = np.nan_to_num(target)
        return f0

    def get_f0_hybrid(
        self,
        methods,
        x,
        f0_min,
        f0_max,
        p_len,
        hop_length,
    ):
        """
        Estimates the fundamental frequency (F0) using a hybrid approach combining multiple methods.

        Args:
            methods: A set of strings specifying the methods to combine.
            x: The input audio signal as a NumPy array.
            f0_min: Minimum F0 value to consider.
            f0_max: Maximum F0 value to consider.
            p_len: Desired length of the F0 output.
            hop_length: Hop length for F0 estimation methods.

        """
        f0_computation_stack = []
        logger.info(
            "Calculating f0 pitch estimations for methods: %s",
            ", ".join(methods),
        )
        # x = x.astype(np.float32)
        # x /= np.quantile(np.abs(x), 0.999)
        for method in methods:
            f0 = None
            if method == "crepe":
                f0 = self.get_f0_crepe(
                    x,
                    f0_min,
                    f0_max,
                    p_len,
                    int(hop_length),
                )
            elif method == "crepe-tiny":
                f0 = self.get_f0_crepe(
                    x,
                    self.f0_min,
                    self.f0_max,
                    p_len,
                    int(hop_length),
                    "tiny",
                )
            elif method == "rmvpe":
                f0 = self.model_rmvpe.infer_from_audio(x, thred=0.03)
                f0 = f0[1:]
            elif method == "fcpe":
                self.model_fcpe = FCPEF0Predictor(
                    os.path.join(str(RVC_MODELS_DIR), "predictors", "fcpe.pt"),
                    f0_min=int(f0_min),
                    f0_max=int(f0_max),
                    dtype=torch.float32,
                    device=self.device,
                    sample_rate=self.sample_rate,
                    threshold=0.03,
                )
                f0 = self.model_fcpe.compute_f0(x, p_len=p_len)
                del self.model_fcpe
                gc.collect()
            f0_computation_stack.append(f0)

        f0_computation_stack = [fc for fc in f0_computation_stack if fc is not None]
        f0_median_hybrid = None
        if len(f0_computation_stack) == 1:
            f0_median_hybrid = f0_computation_stack[0]
        else:
            f0_median_hybrid = np.nanmedian(f0_computation_stack, axis=0)
        return f0_median_hybrid

    def get_f0(
        self,
        input_audio_path,
        x,
        p_len,
        pitch,
        f0_methods,
        hop_length,
        f0_autotune,
        f0_autotune_strength,
        inp_f0=None,
    ):
        """
        Estimates the fundamental frequency (F0) of a given audio signal using various methods.

        Args:
            input_audio_path: Path to the input audio file.
            x: The input audio signal as a NumPy array.
            p_len: Desired length of the F0 output.
            pitch: Key to adjust the pitch of the F0 contour.
            f0_methods: list of methods to use for F0 estimation (e.g., "[crepe]").
            hop_length: Hop length for F0 estimation methods.
            f0_autotune: Whether to apply autotune to the F0 contour.
            inp_f0: Optional input F0 contour to use instead of estimating.

        """
        global input_audio_path2wav
        # input_audio_path2wav[input_audio_path] = x.astype(np.double)
        f0 = self.get_f0_hybrid(
            f0_methods,
            x,
            self.f0_min,
            self.f0_max,
            p_len,
            hop_length,
        )

        if f0_autotune is True:
            f0 = Autotune.autotune_f0(self, f0, f0_autotune_strength)

        f0 *= pow(2, pitch / 12)
        tf0 = self.sample_rate // self.window
        if inp_f0 is not None:
            delta_t = np.round(
                (inp_f0[:, 0].max() - inp_f0[:, 0].min()) * tf0 + 1,
            ).astype("int16")
            replace_f0 = np.interp(
                list(range(delta_t)),
                inp_f0[:, 0] * 100,
                inp_f0[:, 1],
            )
            shape = f0[self.x_pad * tf0 : self.x_pad * tf0 + len(replace_f0)].shape[0]
            f0[self.x_pad * tf0 : self.x_pad * tf0 + len(replace_f0)] = replace_f0[
                :shape
            ]
        f0bak = f0.copy()
        f0_mel = 1127 * np.log(1 + f0 / 700)
        f0_mel[f0_mel > 0] = (f0_mel[f0_mel > 0] - self.f0_mel_min) * 254 / (
            self.f0_mel_max - self.f0_mel_min
        ) + 1
        f0_mel[f0_mel <= 1] = 1
        f0_mel[f0_mel > 255] = 255
        f0_coarse = np.rint(f0_mel).astype(int)

        return f0_coarse, f0bak

    def voice_conversion(
        self,
        model,
        net_g,
        sid,
        audio0,
        pitch,
        pitchf,
        index,
        big_npy,
        index_rate,
        version,
        protect,
    ):
        """
        Performs voice conversion on a given audio segment.

        Args:
            model: The feature extractor model.
            net_g: The generative model for synthesizing speech.
            sid: Speaker ID for the target voice.
            audio0: The input audio segment.
            pitch: Quantized F0 contour for pitch guidance.
            pitchf: Original F0 contour for pitch guidance.
            index: FAISS index for speaker embedding retrieval.
            big_npy: Speaker embeddings stored in a NumPy array.
            index_rate: Blending rate for speaker embedding retrieval.
            version: Model version (Keep to support old models)..
            protect: Protection level for preserving the original pitch.

        """
        with torch.no_grad():
            pitch_guidance = pitch != None and pitchf != None
            # prepare source audio
            feats = torch.from_numpy(audio0).float()
            feats = feats.mean(-1) if feats.dim() == 2 else feats
            assert feats.dim() == 1, feats.dim()
            feats = feats.view(1, -1).to(self.device)
            # extract features
            feats = model(feats)["last_hidden_state"]
            feats = (
                model.final_proj(feats[0]).unsqueeze(0) if version == "v1" else feats
            )
            # make a copy for pitch guidance and protection
            feats0 = feats.clone() if pitch_guidance else None
            if (
                index
            ):  # set by parent function, only true if index is available, loaded, and index rate > 0
                feats = self._retrieve_speaker_embeddings(
                    feats,
                    index,
                    big_npy,
                    index_rate,
                )
            # feature upsampling
            feats = F.interpolate(feats.permute(0, 2, 1), scale_factor=2).permute(
                0,
                2,
                1,
            )
            # adjust the length if the audio is short
            p_len = min(audio0.shape[0] // self.window, feats.shape[1])
            if pitch_guidance:
                feats0 = F.interpolate(feats0.permute(0, 2, 1), scale_factor=2).permute(
                    0,
                    2,
                    1,
                )
                pitch, pitchf = pitch[:, :p_len], pitchf[:, :p_len]
                # Pitch protection blending
                if protect < 0.5:
                    pitchff = pitchf.clone()
                    pitchff[pitchf > 0] = 1
                    pitchff[pitchf < 1] = protect
                    feats = feats * pitchff.unsqueeze(-1) + feats0 * (
                        1 - pitchff.unsqueeze(-1)
                    )
                    feats = feats.to(feats0.dtype)
            else:
                pitch, pitchf = None, None
            p_len = torch.tensor([p_len], device=self.device).long()
            audio1 = (
                (net_g.infer(feats.float(), p_len, pitch, pitchf.float(), sid)[0][0, 0])
                .data.cpu()
                .float()
                .numpy()
            )
            # clean up
            del feats, feats0, p_len
            if torch.cuda.is_available():
                torch.cuda.empty_cache()
        return audio1

    def _retrieve_speaker_embeddings(self, feats, index, big_npy, index_rate):
        npy = feats[0].cpu().numpy()
        score, ix = index.search(npy, k=8)
        weight = np.square(1 / score)
        weight /= weight.sum(axis=1, keepdims=True)
        npy = np.sum(big_npy[ix] * np.expand_dims(weight, axis=2), axis=1)
        feats = (
            torch.from_numpy(npy).unsqueeze(0).to(self.device) * index_rate
            + (1 - index_rate) * feats
        )
        return feats

    def pipeline(
        self,
        model,
        net_g,
        sid,
        audio,
        pitch,
        f0_methods,
        file_index,
        index_rate,
        pitch_guidance,
        volume_envelope,
        version,
        protect,
        hop_length,
        f0_autotune,
        f0_autotune_strength,
        f0_file,
    ):
        """
        The main pipeline function for performing voice conversion.

        Args:
            model: The feature extractor model.
            net_g: The generative model for synthesizing speech.
            sid: Speaker ID for the target voice.
            audio: The input audio signal.
            input_audio_path: Path to the input audio file.
            pitch: Key to adjust the pitch of the F0 contour.
            f0_methods: Methods to use for F0 estimation.
            file_index: Path to the FAISS index file for speaker embedding retrieval.
            index_rate: Blending rate for speaker embedding retrieval.
            pitch_guidance: Whether to use pitch guidance during voice conversion.
            tgt_sr: Target sampling rate for the output audio.
            resample_sr: Resampling rate for the output audio.
            volume_envelope: Blending rate for adjusting the RMS level of the output audio.
            version: Model version.
            protect: Protection level for preserving the original pitch.
            hop_length: Hop length for F0 estimation methods.
            f0_autotune: Whether to apply autotune to the F0 contour.
            f0_file: Path to a file containing an F0 contour to use.

        """
        if file_index != "" and os.path.exists(file_index) and index_rate > 0:
            try:
                index = faiss.read_index(file_index)
                big_npy = index.reconstruct_n(0, index.ntotal)
            except Exception as error:
                print(f"An error occurred reading the FAISS index: {error}")
                index = big_npy = None
        else:
            index = big_npy = None
        audio = signal.filtfilt(bh, ah, audio)
        audio_pad = np.pad(audio, (self.window // 2, self.window // 2), mode="reflect")
        opt_ts = []
        if audio_pad.shape[0] > self.t_max:
            audio_sum = np.zeros_like(audio)
            for i in range(self.window):
                audio_sum += audio_pad[i : i - self.window]
            for t in range(self.t_center, audio.shape[0], self.t_center):
                opt_ts.append(
                    t
                    - self.t_query
                    + np.where(
                        np.abs(audio_sum[t - self.t_query : t + self.t_query])
                        == np.abs(audio_sum[t - self.t_query : t + self.t_query]).min(),
                    )[0][0],
                )
        s = 0
        audio_opt = []
        t = None
        audio_pad = np.pad(audio, (self.t_pad, self.t_pad), mode="reflect")
        p_len = audio_pad.shape[0] // self.window
        inp_f0 = None
        if hasattr(f0_file, "name"):
            try:
                with open(f0_file.name) as f:
                    lines = f.read().strip("\n").split("\n")
                inp_f0 = []
                for line in lines:
                    inp_f0.append([float(i) for i in line.split(",")])
                inp_f0 = np.array(inp_f0, dtype="float32")
            except Exception as error:
                print(f"An error occurred reading the F0 file: {error}")
        sid = torch.tensor(sid, device=self.device).unsqueeze(0).long()
        if pitch_guidance:
            pitch, pitchf = self.get_f0(
                "input_audio_path",  # questionable purpose of making a key for an array
                audio_pad,
                p_len,
                pitch,
                f0_methods,
                hop_length,
                f0_autotune,
                f0_autotune_strength,
                inp_f0,
            )
            pitch = pitch[:p_len]
            pitchf = pitchf[:p_len]
            if self.device == "mps":
                pitchf = pitchf.astype(np.float32)
            pitch = torch.tensor(pitch, device=self.device).unsqueeze(0).long()
            pitchf = torch.tensor(pitchf, device=self.device).unsqueeze(0).float()
        for t in opt_ts:
            t = t // self.window * self.window
            if pitch_guidance:
                audio_opt.append(
                    self.voice_conversion(
                        model,
                        net_g,
                        sid,
                        audio_pad[s : t + self.t_pad2 + self.window],
                        pitch[:, s // self.window : (t + self.t_pad2) // self.window],
                        pitchf[:, s // self.window : (t + self.t_pad2) // self.window],
                        index,
                        big_npy,
                        index_rate,
                        version,
                        protect,
                    )[self.t_pad_tgt : -self.t_pad_tgt],
                )
            else:
                audio_opt.append(
                    self.voice_conversion(
                        model,
                        net_g,
                        sid,
                        audio_pad[s : t + self.t_pad2 + self.window],
                        None,
                        None,
                        index,
                        big_npy,
                        index_rate,
                        version,
                        protect,
                    )[self.t_pad_tgt : -self.t_pad_tgt],
                )
            s = t
        if pitch_guidance:
            audio_opt.append(
                self.voice_conversion(
                    model,
                    net_g,
                    sid,
                    audio_pad[t:],
                    pitch[:, t // self.window :] if t is not None else pitch,
                    pitchf[:, t // self.window :] if t is not None else pitchf,
                    index,
                    big_npy,
                    index_rate,
                    version,
                    protect,
                )[self.t_pad_tgt : -self.t_pad_tgt],
            )
        else:
            audio_opt.append(
                self.voice_conversion(
                    model,
                    net_g,
                    sid,
                    audio_pad[t:],
                    None,
                    None,
                    index,
                    big_npy,
                    index_rate,
                    version,
                    protect,
                )[self.t_pad_tgt : -self.t_pad_tgt],
            )
        audio_opt = np.concatenate(audio_opt)
        if volume_envelope != 1:
            audio_opt = AudioProcessor.change_rms(
                audio,
                self.sample_rate,
                audio_opt,
                self.sample_rate,
                volume_envelope,
            )
        audio_max = np.abs(audio_opt).max() / 0.99
        if audio_max > 1:
            audio_opt /= audio_max
        if pitch_guidance:
            del pitch, pitchf
        del sid
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
        return audio_opt



"""Module which defines functions used for managing various settings."""

from __future__ import annotations

from ultimate_rvc.common import TEMP_DIR
from ultimate_rvc.core.manage.common import delete_directory


def delete_temp_files() -> None:
    """Delete all temporary files."""
    delete_directory(TEMP_DIR)



"""
Common definitions for modules in the Ultimate RVC project that
generate audio.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import lazy_loader as lazy

import logging
from functools import cache, reduce
from pathlib import Path

from rich import print as rprint

from ultimate_rvc.core.common import (
    get_file_hash,
    get_hash,
    json_dump,
    json_dumps,
    json_load,
    validate_audio_dir_exists,
    validate_audio_file_exists,
    validate_model,
)
from ultimate_rvc.core.exceptions import (
    Entity,
    NotFoundError,
    NotProvidedError,
    UIMessage,
)
from ultimate_rvc.core.generate.typing_extra import (
    AudioExtInternal,
    FileMetaData,
    MixedAudioMetaData,
    MixedAudioType,
    RVCAudioMetaData,
    StagedAudioMetaData,
    WaveifiedAudioMetaData,
)
from ultimate_rvc.typing_extra import (
    AudioExt,
    EmbedderModel,
    F0Method,
    RVCContentType,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

    import ffmpeg
    import static_ffmpeg

    # NOTE the only reason this module is imported here is so we can
    # annotate the return value of the _get_voice_converter function.
    from ultimate_rvc.rvc.infer.infer import VoiceConverter
    from ultimate_rvc.typing_extra import Json, StrPath
else:
    static_ffmpeg = lazy.load("static_ffmpeg")
    ffmpeg = lazy.load("ffmpeg")
logger = logging.getLogger(__name__)


# NOTE consider increasing hash_size to 16. Otherwise
# we might have problems with hash collisions when using app as CLI
def get_unique_base_path(
    directory: StrPath,
    prefix: str,
    args_dict: Json,
    hash_size: int = 5,
) -> Path:
    """
    Get a unique base path (a path without any extension) for a file in
    a directory by hashing the arguments used to generate
    the audio that is stored or will be stored in that file.

    Parameters
    ----------
    directory :StrPath
        The path to a directory.
    prefix : str
        The prefix to use for the base path.
    args_dict : Json
        A JSON-serializable dictionary of named arguments used to
        generate the audio that is stored or will be stored in a file
        in the given directory.
    hash_size : int, default=5
        The size (in bytes) of the hash to use for the base path.

    Returns
    -------
    Path
        The unique base path for a file in a song directory.

    """
    directory_path = Path(directory)
    dict_hash = get_hash(args_dict, size=hash_size)
    while True:
        base_path = directory_path / f"{prefix}_{dict_hash}"
        json_path = base_path.with_suffix(".json")
        if json_path.exists():
            file_dict = json_load(json_path)
            if file_dict == args_dict:
                return base_path
            dict_hash = get_hash(dict_hash, size=hash_size)
            rprint("[~] Rehashing...")
        else:
            return base_path


def wavify(
    audio_track: StrPath,
    directory: StrPath,
    prefix: str,
    accepted_formats: set[AudioExt] | None = None,
) -> Path:
    """
    Convert a given audio track to wav format if its current format is
    one of the given accepted formats.

    Parameters
    ----------
    audio_track : StrPath
        The path to the audio track to convert.
    directory : StrPath
        The path to the directory where the converted audio track
        will be saved.
    prefix : str
        The prefix to use for the name of the converted audio track.
    accepted_formats : set[AudioExt], optional
        The audio formats to accept for conversion. If None, the
        accepted formats are mp3, ogg, flac, m4a and aac.

    Returns
    -------
    Path
        The path to the audio track in wav format or the original audio
        track if it is not in one of the accepted formats.

    """
    static_ffmpeg.add_paths(weak=True)

    audio_path = validate_audio_file_exists(audio_track, Entity.AUDIO_TRACK)
    dir_path = validate_audio_dir_exists(directory, Entity.DIRECTORY)

    wav_path = audio_path

    if accepted_formats is None:
        accepted_formats = set(AudioExt) - {AudioExt.WAV}

    # NOTE The lazy_import function does not work with pydub
    # so we import it here manually
    import pydub.utils as pydub_utils  # noqa: PLC0415

    audio_info = pydub_utils.mediainfo(str(audio_path))
    logger.info("Audio info:\n%s", json_dumps(audio_info))
    if any(
        (
            accepted_format in audio_info["format_name"]
            if accepted_format == AudioExt.M4A
            else accepted_format == audio_info["format_name"]
        )
        for accepted_format in accepted_formats
    ):
        args_dict = WaveifiedAudioMetaData(
            audio_track=FileMetaData(
                name=audio_path.name,
                hash_id=get_file_hash(audio_path),
            ),
        ).model_dump()

        paths = [
            get_unique_base_path(
                dir_path,
                prefix,
                args_dict,
            ).with_suffix(suffix)
            for suffix in [".wav", ".json"]
        ]
        wav_path, wav_json_path = paths
        if not all(path.exists() for path in paths):
            _, stderr = (
                ffmpeg.input(audio_path)
                .output(filename=wav_path, f="wav")
                .run(
                    overwrite_output=True,
                    quiet=True,
                )
            )
            logger.info("FFmpeg stderr:\n%s", stderr.decode("utf-8"))
            json_dump(args_dict, wav_json_path)

    return wav_path


def _get_rvc_files(model_name: str) -> tuple[Path, Path | None]:
    """
    Get the RVC model file and potential index file of a voice model.

    Parameters
    ----------
    model_name : str
        The name of the voice model to get the RVC files of.

    Returns
    -------
    model_file : Path
        The path to the RVC model file.
    index_file : Path | None
        The path to the RVC index file, if it exists.

    Raises
    ------
    NotFoundError
        If no model file exists in the voice model directory.


    """
    model_dir_path = validate_model(model_name, Entity.VOICE_MODEL)
    file_path_map = {
        ext: path
        for path in model_dir_path.iterdir()
        for ext in [".pth", ".index"]
        if ext == path.suffix
    }

    if ".pth" not in file_path_map:
        raise NotFoundError(
            entity=Entity.MODEL_FILE,
            location=model_dir_path,
            is_path=False,
        )

    model_file = file_path_map[".pth"]
    index_file = file_path_map.get(".index")
    return model_file, index_file


@cache
def _get_voice_converter() -> VoiceConverter:
    """
    Get a voice converter.

    Returns
    -------
    VoiceConverter
        A voice converter.

    """
    from ultimate_rvc.rvc.infer.infer import VoiceConverter  # noqa: PLC0415

    return VoiceConverter()


def convert(
    audio_track: StrPath,
    directory: StrPath,
    model_name: str,
    n_octaves: int = 0,
    n_semitones: int = 0,
    f0_methods: Sequence[F0Method] | None = None,
    index_rate: float = 0.3,
    rms_mix_rate: float = 1.0,
    protect_rate: float = 0.33,
    hop_length: int = 128,
    split_audio: bool = False,
    autotune_audio: bool = False,
    autotune_strength: float = 1.0,
    clean_audio: bool = False,
    clean_strength: float = 0.7,
    embedder_model: EmbedderModel = EmbedderModel.CONTENTVEC,
    custom_embedder_model: str | None = None,
    sid: int = 0,
    content_type: RVCContentType = RVCContentType.AUDIO,
    make_directory: bool = False,
) -> Path:
    """
    Convert an audio track using an RVC model.

    Parameters
    ----------
    audio_track : StrPath
        The path to the audio track to convert.
    directory : StrPath
        The path to the directory where the converted audio track
        will be saved.
    model_name : str
        The name of the model to use for voice conversion.
    n_octaves : int, default=0
        The number of octaves to pitch-shift the converted audio by.
    n_semitones : int, default=0
        The number of semitones to pitch-shift the converted audio by.
    f0_methods : Sequence[F0Method], optional
        The methods to use for pitch extraction. If None, the method
        used is rmvpe.
    index_rate : float, default=0.3
        The influence of the index file on the voice conversion.
    rms_mix_rate : float, default = 1.0
        The blending rate of the volume envelope of the converted
        audio.
    protect_rate : float, default=0.33
        The protection rate for consonants and breathing sounds.
    hop_length : int, default=128
        The hop length to use for CREPE-based pitch extraction.
    split_audio : bool, default=False
        Whether to split the audio track into smaller segments before
        converting it.
    autotune_audio : bool, default=False
        Whether to apply autotune to the converted audio.
    autotune_strength : float, default=1.0
        The strength of the autotune to apply to the converted audio.
    clean_audio : bool, default=False
        Whether to clean the converted audio.
    clean_strength : float, default=0.7
        The intensity of the cleaning to apply to the converted audio.
    embedder_model : EmbedderModel, default=EmbedderModel.CONTENTVEC
        The model to use for generating speaker embeddings.
    custom_embedder_model : str, optional
        The name of a custom embedder model to use for generating
        speaker embeddings.
    sid : int, default=0
        The speaker id to use for multi-speaker models.
    content_type : RVCContentType, default=RVCContentType.AUDIO
        The type of content to convert. Determines what is shown in
        display mesages and saved file names.
    make_directory : bool, default=False
        Whether to create the directory where the converted audio
        track will be saved if it does not exist.

    Returns
    -------
    Path
        The path to the converted audio track.

    """
    match content_type:
        case RVCContentType.VOCALS:
            track_entity = Entity.VOCALS_TRACK
            directory_entity = Entity.SONG_DIR
        case RVCContentType.VOICE:
            track_entity = Entity.VOICE_TRACK
            directory_entity = Entity.DIRECTORY
        case RVCContentType.SPEECH:
            track_entity = Entity.SPEECH_TRACK
            directory_entity = Entity.DIRECTORY
        case RVCContentType.AUDIO:
            track_entity = Entity.AUDIO_TRACK
            directory_entity = Entity.DIRECTORY
    audio_path = validate_audio_file_exists(audio_track, track_entity)
    if make_directory:
        Path(directory).mkdir(parents=True, exist_ok=True)
    directory_path = validate_audio_dir_exists(directory, directory_entity)
    validate_model(model_name, Entity.VOICE_MODEL)
    custom_embedder_model_path = None
    if embedder_model == EmbedderModel.CUSTOM:
        custom_embedder_model_path = validate_model(
            custom_embedder_model,
            Entity.CUSTOM_EMBEDDER_MODEL,
        )

    audio_path = wavify(
        audio_path,
        directory_path,
        "20_Input",
        accepted_formats={AudioExt.M4A, AudioExt.AAC},
    )

    n_semitones = n_octaves * 12 + n_semitones
    f0_methods_set = set(f0_methods) if f0_methods else {F0Method.RMVPE}

    args_dict = RVCAudioMetaData(
        audio_track=FileMetaData(
            name=audio_path.name,
            hash_id=get_file_hash(audio_path),
        ),
        model_name=model_name,
        n_semitones=n_semitones,
        f0_methods=sorted(f0_methods_set),
        index_rate=index_rate,
        rms_mix_rate=rms_mix_rate,
        protect_rate=protect_rate,
        hop_length=hop_length,
        split_audio=split_audio,
        autotune_audio=autotune_audio,
        autotune_strength=autotune_strength,
        clean_audio=clean_audio,
        clean_strength=clean_strength,
        embedder_model=embedder_model,
        custom_embedder_model=custom_embedder_model,
        sid=sid,
    ).model_dump()

    paths = [
        get_unique_base_path(
            directory_path,
            f"21_{content_type.capitalize()}_Converted",
            args_dict,
        ).with_suffix(suffix)
        for suffix in [".wav", ".json"]
    ]

    converted_audio_path, converted_audio_json_path = paths

    if not all(path.exists() for path in paths):
        rvc_model_path, rvc_index_path = _get_rvc_files(model_name)

        voice_converter = _get_voice_converter()

        voice_converter.convert_audio(
            audio_input_path=str(audio_path),
            audio_output_path=str(converted_audio_path),
            model_path=str(rvc_model_path),
            index_path=str(rvc_index_path) if rvc_index_path else "",
            pitch=n_semitones,
            f0_methods=f0_methods_set,
            index_rate=index_rate,
            volume_envelope=rms_mix_rate,
            protect=protect_rate,
            hop_length=hop_length,
            split_audio=split_audio,
            f0_autotune=autotune_audio,
            f0_autotune_strength=autotune_strength,
            embedder_model=embedder_model,
            embedder_model_custom=(
                str(custom_embedder_model_path)
                if custom_embedder_model_path is not None
                else None
            ),
            clean_audio=clean_audio,
            clean_strength=clean_strength,
            post_process=False,
            resample_sr=0,
            sid=sid,
        )
        json_dump(args_dict, converted_audio_json_path)
    return converted_audio_path


def _to_internal(audio_ext: AudioExt) -> AudioExtInternal:
    """
    Map an audio extension to an internally recognized format.

    Parameters
    ----------
    audio_ext : AudioExt
        The audio extension to map.

    Returns
    -------
    AudioExtInternal
        The internal audio extension.

    """
    match audio_ext:
        case AudioExt.M4A:
            return AudioExtInternal.IPOD
        case AudioExt.AAC:
            return AudioExtInternal.ADTS
        case _:
            return AudioExtInternal(audio_ext)


def _mix_audio(
    audio_track_gain_pairs: Sequence[tuple[StrPath, int]],
    output_file: StrPath,
    output_sr: int = 44100,
    output_format: AudioExt = AudioExt.MP3,
) -> None:
    """
    Mix multiple audio tracks.

    Parameters
    ----------
    audio_track_gain_pairs : Sequence[tuple[StrPath, int]]
        A sequence of pairs each containing the path to an audio track
        and the gain to apply to it.
    output_file : StrPath
        The path to the file to save the mixed audio to.
    output_sr : int, default=44100
        The sample rate of the mixed audio.
    output_format : AudioExt, default=AudioExt.MP3
        The audio format of the mixed audio.

    """
    static_ffmpeg.add_paths(weak=True)
    # NOTE The lazy_import function does not work with pydub
    # so we import it here manually
    import pydub  # noqa: PLC0415

    mixed_audio = reduce(
        lambda a1, a2: a1.overlay(a2),
        [
            pydub.AudioSegment.from_wav(audio_track) + gain
            for audio_track, gain in audio_track_gain_pairs
        ],
    )
    mixed_audio_resampled = mixed_audio.set_frame_rate(output_sr)
    mixed_audio_resampled.export(
        output_file,
        format=_to_internal(output_format),
    )


def mix_audio(
    audio_track_gain_pairs: Sequence[tuple[StrPath, int]],
    directory: StrPath,
    output_sr: int = 44100,
    output_format: AudioExt = AudioExt.MP3,
    content_type: MixedAudioType = MixedAudioType.AUDIO,
) -> Path:
    """
    Mix one or more audio tracks.

    Parameters
    ----------
    audio_track_gain_pairs : Sequence[tuple[StrPath, int]]
        A sequence of pairs each containing the path to an audio track
        and the gain to apply to it.
    directory : StrPath
        The path to the directory where the mixed audio will be saved.
    output_sr : int, default=44100
        The sample rate of the mixed audio.
    output_format : AudioExt, default=AudioExt.MP3
        The audio format of the mixed audio.
    content_type: MixedAudioType, default=MixedAudioType.AUDIO
        The type of mixed audio. Determines what is shown in display
        messages and saved file names.

    Returns
    -------
    Path
        The path to the mixed audio.

    Raises
    ------
    NotProvidedError
        If no audio tracks are provided.

    """
    match content_type:
        case MixedAudioType.AUDIO:
            directory_entity = Entity.DIRECTORY
        case MixedAudioType.SPEECH:
            directory_entity = Entity.DIRECTORY
        case MixedAudioType.SONG:
            directory_entity = Entity.SONG_DIR
    if not audio_track_gain_pairs:
        raise NotProvidedError(
            entity=Entity.AUDIO_TRACK_GAIN_PAIRS,
            ui_msg=UIMessage.NO_AUDIO_TRACK,
        )

    audio_path_gain_pairs = [
        (
            wavify(
                validate_audio_file_exists(audio_track, Entity.AUDIO_TRACK),
                directory,
                "50_Input",
            ),
            gain,
        )
        for audio_track, gain in audio_track_gain_pairs
    ]
    dir_path = validate_audio_dir_exists(directory, directory_entity)
    args_dict = MixedAudioMetaData(
        staged_audio_tracks=[
            StagedAudioMetaData(
                audio_track=FileMetaData(
                    name=audio_path.name,
                    hash_id=get_file_hash(audio_path),
                ),
                gain=gain,
            )
            for audio_path, gain in audio_path_gain_pairs
        ],
        output_sr=output_sr,
        output_format=output_format,
    ).model_dump()

    paths = [
        get_unique_base_path(
            dir_path,
            f"51_{content_type.capitalize()}_Mixed",
            args_dict,
        ).with_suffix(suffix)
        for suffix in ["." + output_format, ".json"]
    ]

    mix_path, mix_json_path = paths

    if not all(path.exists() for path in paths):
        _mix_audio(audio_path_gain_pairs, mix_path, output_sr, output_format)
        json_dump(args_dict, mix_json_path)
    return mix_path



"""
Module which exposes functionality for training voice conversion
models.
"""

from __future__ import annotations

import logging
import os
import re
import signal

from ultimate_rvc.core.common import (
    TRAINING_MODELS_DIR,
    VOICE_MODELS_DIR,
    copy_files_to_new_dir,
    json_dump,
    json_load,
    validate_model,
)
from ultimate_rvc.core.exceptions import (
    Entity,
    IncompatiblePretrainedModelError,
    IncompatibleVocoderError,
    ModelAsssociatedEntityNotFoundError,
    ModelExistsError,
    NotProvidedError,
    Step,
)
from ultimate_rvc.core.train.common import validate_devices
from ultimate_rvc.core.train.typing_extra import ModelInfo, TrainingInfo
from ultimate_rvc.typing_extra import (
    DeviceType,
    IndexAlgorithm,
    PretrainedType,
    Vocoder,
)

logger = logging.getLogger(__name__)


def _get_pretrained_model(
    pretrained_type: PretrainedType,
    vocoder: Vocoder,
    sample_rate: int,
    custom_pretrained: str | None = None,
) -> tuple[str, str]:
    """
    Get the pretrained model to finetune a voice model on.

    Parameters
    ----------
    pretrained_type : PretrainedType
        The type of pretrained model to finetune the voice model on
    vocoder : str
        The vocoder to use for audio synthesis when training the voice
        model.
    sample_rate : int
        The sample rate of the preprocessed dataset associated with the
        voice model to be trained.
    custom_pretrained : str, optional
        The name of a custom pretrained model to finetune the voice
        model on

    Returns
    -------
    pg : str
        The path to the generator of the pretrained model to finetune.
    pd : str
        The path to the discriminator of the pretrained model to
        finetune.

    Raises
    ------
    ModelAsssociatedEntityNotFoundError
        If the voice model to be trained does not have an associated
        dataset file list or if a custom pretrained
        generator/discriminator model does not have an associated
        generator or discriminator.
    IncompatiblePretrainedModelError
        if a custom pretrained model is not compatible with the sample
        rate of the preprocessed dataset associated with the voice model
        to be trained.
    IncompatibleVocoderError
        If the default pretrained model is not compatible with the
        vocoder to be used for audio synthesis when training the voice
        model.

    """
    match pretrained_type:
        case PretrainedType.NONE:
            pg, pd = "", ""
        case PretrainedType.DEFAULT:
            from ultimate_rvc.rvc.lib.tools.pretrained_selector import (  # noqa: PLC0415
                pretrained_selector,
            )

            if not vocoder == Vocoder.HIFI_GAN:
                raise IncompatibleVocoderError(vocoder)

            pg, pd = pretrained_selector(vocoder, sample_rate=sample_rate)
        case PretrainedType.CUSTOM:
            custom_pretrained_path = validate_model(
                custom_pretrained,
                Entity.CUSTOM_PRETRAINED_MODEL,
            )
            # NOTE simply done to appease the type checker
            custom_pretrained = custom_pretrained_path.name

            # TODO need to make this cleaner
            custom_pretrained_sample_rate = int(custom_pretrained[-3:-1]) * 1000
            if not custom_pretrained_sample_rate == sample_rate:
                raise IncompatiblePretrainedModelError(custom_pretrained, sample_rate)

            pg = next(
                (
                    str(path)
                    for path in custom_pretrained_path.iterdir()
                    if re.match(r"^(G|f0G).*\.pth$|.*G\.pth$", path.name)
                ),
                None,
            )
            if pg is None:
                raise ModelAsssociatedEntityNotFoundError(
                    Entity.GENERATOR,
                    custom_pretrained,
                )
            pd = next(
                (
                    str(path)
                    for path in custom_pretrained_path.iterdir()
                    if re.match(r"^(D|f0D).*\.pth$|.*D\.pth$", path.name)
                ),
                None,
            )
            if pd is None:
                raise ModelAsssociatedEntityNotFoundError(
                    Entity.DISCRIMINATOR,
                    custom_pretrained,
                )

    return pg, pd


def run_training(
    model_name: str,
    num_epochs: int = 500,
    batch_size: int = 8,
    detect_overtraining: bool = False,
    overtraining_threshold: int = 50,
    vocoder: Vocoder = Vocoder.HIFI_GAN,
    index_algorithm: IndexAlgorithm = IndexAlgorithm.AUTO,
    pretrained_type: PretrainedType = PretrainedType.DEFAULT,
    custom_pretrained: str | None = None,
    save_interval: int = 10,
    save_all_checkpoints: bool = False,
    save_all_weights: bool = False,
    clear_saved_data: bool = False,
    upload_model: bool = False,
    upload_name: str | None = None,
    hardware_acceleration: DeviceType = DeviceType.AUTOMATIC,
    gpu_ids: set[int] | None = None,
    preload_dataset: bool = False,
    reduce_memory_usage: bool = False,
) -> list[str] | None:
    """

    Train a voice model using its associated preprocessed dataset and
    extracted features.

    Parameters
    ----------
    model_name : str
        The name of the voice model to train.
    num_epochs : int, default=500
        The number of epochs to train the voice model. A higher number
        can improve voice model performance but may lead to
        overtraining.
    batch_size : int, default=8
        The number of samples to include in each training batch. It is
        advisable to align this value with the available VRAM of your
        GPU. A setting of 4 offers improved accuracy but slower
        processing, while 8 provides faster and standard results.
    detect_overtraining : bool, default=False
        Whether to detect overtraining to prevent the voice model from
        learning the training data too well and losing the ability to
        generalize to new data.
    overtraining_threshold : int, default=50
        The maximum number of epochs to continue training without any
        observed improvement in voice model performance.
    vocoder : Vocoder, default=Vocoder.HIFI_GAN
        The vocoder to use for audio synthesis during training. HiFi-GAN
        provides basic audio fidelity, while RefineGAN provides the
        highest audio fidelity.
    index_algorithm : IndexAlgorithm, default=IndexAlgorithm.AUTO
        The method to use for generating an index file for the trained
        voice model. KMeans is particularly useful for large datasets.
    pretrained_type : PretrainedType, default=PretrainedType.DEFAULT
        The type of pretrained model to finetune the voice model on.
        "None" will train the voice model from scratch, while
        "Default" will use a pretrained model tailored to the specific
        voice model architecture. "Custom" will use a custom pretrained
        model that you provide.
    custom_pretrained: str, optional
        The name of a custom pretrained model to finetune the voice
        model on.
    save_interval : int, default=10
        The epoch interval at which to to save voice model weights and
        checkpoints. The best model weights are always saved regardless
        of this setting.
    save_all_checkpoints : bool, default=False
        Whether to save a unique checkpoint at each save interval. If
        not enabled, only the latest checkpoint will be saved at each
        interval.
    save_all_weights : bool, default=False
        Whether to save unique voice model weights at each save
        interval. If not enabled, only the best voice model weights will
        be saved.
    clear_saved_data : bool, default=False
        Whether to delete any existing training data associated
        with the voice model before training commences. Enable this
        setting only if you are training a new voice model from scratch
        or restarting training.
    upload_model : bool, default=False
        Whether to automatically upload the trained voice model so that
        it can be used for audio generation tasks within the Ultimate
        RVC app.
    upload_name : str, optional
        The name to give the uploaded voice model.
    hardware_acceleration : DeviceType, default=DeviceType.AUTOMATIC
        The type of hardware acceleration to use when training the voice
        model. `AUTOMATIC` will select the first available GPU and fall
        back to CPU if no GPUs are available.
    gpu_ids : set[int], optional
        Set of ids of the GPUs to use for training the voice model when
        `GPU` is selected for hardware acceleration.
    preload_dataset : bool, default=False
        Whether to preload all training data into GPU memory. This can
        improve training speed but requires a lot of VRAM.
    reduce_memory_usage : bool, default=False
        Whether to reduce VRAM usage at the cost of slower training
        speed by enabling activation checkpointing. This is useful for
        GPUs with limited memory (e.g., <6GB VRAM) or when training with
        a batch size larger than what your GPU can normally accommodate.

    Returns
    -------
    list[str] | None
        A list containing the paths to the best weights file and the
        index file for the trained voice model, if they exist.
        Otherwise, None.

    Raises
    ------
    ModelAsssociatedEntityNotFoundError
        If the voice model to be trained does not have an associated
        dataset file list.
    NotProvidedError
        If an upload name is not provided when the upload parameter is
        set
    ModelExistsError
        If a voice with the provided upload name already exists when the
        upload parameter is set


    """
    model_path = validate_model(model_name, Entity.TRAINING_MODEL)
    filelist_path = model_path / "filelist.txt"
    if not filelist_path.is_file():
        raise ModelAsssociatedEntityNotFoundError(
            Entity.DATASET_FILE_LIST,
            model_name,
            Step.FEATURE_EXTRACTION,
        )
    upload_model_path = None
    if upload_model:
        if not upload_name:
            raise NotProvidedError(Entity.UPLOAD_NAME)
        upload_model_path = VOICE_MODELS_DIR / upload_name.strip()
        if upload_model_path.is_dir():
            raise ModelExistsError(Entity.VOICE_MODEL, upload_name)

    model_info_dict = json_load(model_path / "model_info.json")

    model_info = ModelInfo.model_validate(model_info_dict)
    sample_rate = model_info.sample_rate

    pg, pd = _get_pretrained_model(
        pretrained_type,
        vocoder,
        sample_rate,
        custom_pretrained,
    )

    from ultimate_rvc.rvc.train.train import main as train_main  # noqa: PLC0415

    device_type, device_ids = validate_devices(hardware_acceleration, gpu_ids)

    train_main(
        model_name,
        sample_rate,
        vocoder,
        num_epochs,
        batch_size,
        save_interval,
        not save_all_checkpoints,
        save_all_weights,
        pg,
        pd,
        detect_overtraining,
        overtraining_threshold,
        clear_saved_data,
        preload_dataset,
        reduce_memory_usage,
        device_type,
        device_ids,
    )

    model_file = model_path / f"{model_name}_best.pth"

    if not model_file.is_file():
        return None

    from ultimate_rvc.rvc.train.process.extract_index import (  # noqa: PLC0415
        main as extract_index_main,
    )

    extract_index_main(str(model_path), index_algorithm)

    index_file = model_path / f"{model_name}.index"

    if not index_file.is_file():
        return None
    if upload_model_path:
        copy_files_to_new_dir([index_file, model_file], upload_model_path)
    return [str(model_file), str(index_file)]


def stop_training(model_name: str) -> None:
    """
    Stop the training of a voice model.

    Parameters
    ----------
    model_name : str
        The name of the voice model to stop training for.

    """
    training_info_path = TRAINING_MODELS_DIR / model_name / "config.json"
    try:
        training_info_dict = json_load(training_info_path)
        training_info = TrainingInfo.model_validate(training_info_dict)
        process_ids = training_info.process_pids
        for pid in process_ids:
            os.kill(pid, signal.SIGTERM)
        training_info.process_pids = []
        updated_training_info_dict = training_info.model_dump()
        json_dump(updated_training_info_dict, training_info_path)
    except Exception as e:  # noqa: BLE001
        logger.error("Error stopping training: %s", e)  # noqa: TRY400



import logging
import math

import torch

from ultimate_rvc.rvc.lib.algorithm.attentions import FFN, MultiHeadAttention
from ultimate_rvc.rvc.lib.algorithm.commons import sequence_mask
from ultimate_rvc.rvc.lib.algorithm.modules import WaveNet
from ultimate_rvc.rvc.lib.algorithm.normalization import LayerNorm

logger = logging.getLogger(__name__)


class Encoder(torch.nn.Module):
    """
    Encoder module for the Transformer model.

    Args:
        hidden_channels (int): Number of hidden channels in the encoder.
        filter_channels (int): Number of filter channels in the feed-forward network.
        n_heads (int): Number of attention heads.
        n_layers (int): Number of encoder layers.
        kernel_size (int, optional): Kernel size of the convolution layers in the feed-forward network. Defaults to 1.
        p_dropout (float, optional): Dropout probability. Defaults to 0.0.
        window_size (int, optional): Window size for relative positional encoding. Defaults to 10.

    """

    def __init__(
        self,
        hidden_channels: int,
        filter_channels: int,
        n_heads: int,
        n_layers: int,
        kernel_size: int = 1,
        p_dropout: float = 0.0,
        window_size: int = 10,
    ):
        super().__init__()

        self.hidden_channels = hidden_channels
        self.n_layers = n_layers
        self.drop = torch.nn.Dropout(p_dropout)

        self.attn_layers = torch.nn.ModuleList(
            [
                MultiHeadAttention(
                    hidden_channels,
                    hidden_channels,
                    n_heads,
                    p_dropout=p_dropout,
                    window_size=window_size,
                )
                for _ in range(n_layers)
            ],
        )
        self.norm_layers_1 = torch.nn.ModuleList(
            [LayerNorm(hidden_channels) for _ in range(n_layers)],
        )
        self.ffn_layers = torch.nn.ModuleList(
            [
                FFN(
                    hidden_channels,
                    hidden_channels,
                    filter_channels,
                    kernel_size,
                    p_dropout=p_dropout,
                )
                for _ in range(n_layers)
            ],
        )
        self.norm_layers_2 = torch.nn.ModuleList(
            [LayerNorm(hidden_channels) for _ in range(n_layers)],
        )

    def forward(self, x, x_mask):
        attn_mask = x_mask.unsqueeze(2) * x_mask.unsqueeze(-1)
        x = x * x_mask

        for i in range(self.n_layers):
            y = self.attn_layers[i](x, x, attn_mask)
            y = self.drop(y)
            x = self.norm_layers_1[i](x + y)

            y = self.ffn_layers[i](x, x_mask)
            y = self.drop(y)
            x = self.norm_layers_2[i](x + y)

        return x * x_mask


class TextEncoder(torch.nn.Module):
    """
    Text Encoder with configurable embedding dimension.

    Args:
        out_channels (int): Output channels of the encoder.
        hidden_channels (int): Hidden channels of the encoder.
        filter_channels (int): Filter channels of the encoder.
        n_heads (int): Number of attention heads.
        n_layers (int): Number of encoder layers.
        kernel_size (int): Kernel size of the convolutional layers.
        p_dropout (float): Dropout probability.
        embedding_dim (int): Embedding dimension for phone embeddings (v1 = 256, v2 = 768).
        f0 (bool, optional): Whether to use F0 embedding. Defaults to True.

    """

    def __init__(
        self,
        out_channels: int,
        hidden_channels: int,
        filter_channels: int,
        n_heads: int,
        n_layers: int,
        kernel_size: int,
        p_dropout: float,
        embedding_dim: int,
        f0: bool = True,
    ):
        super().__init__()
        self.hidden_channels = hidden_channels
        self.out_channels = out_channels
        self.emb_phone = torch.nn.Linear(embedding_dim, hidden_channels)
        self.lrelu = torch.nn.LeakyReLU(0.1, inplace=True)
        self.emb_pitch = torch.nn.Embedding(256, hidden_channels) if f0 else None
        logger.info("hidden_channels: %d", hidden_channels)

        self.encoder = Encoder(
            hidden_channels,
            filter_channels,
            n_heads,
            n_layers,
            kernel_size,
            p_dropout,
        )
        self.proj = torch.nn.Conv1d(hidden_channels, out_channels * 2, 1)

    def forward(
        self,
        phone: torch.Tensor,
        pitch: torch.Tensor | None,
        lengths: torch.Tensor,
    ):
        x = self.emb_phone(phone)
        if pitch is not None and self.emb_pitch:
            x += self.emb_pitch(pitch)

        x *= math.sqrt(self.hidden_channels)
        x = self.lrelu(x)
        x = x.transpose(1, -1)  # [B, H, T]

        x_mask = sequence_mask(lengths, x.size(2)).unsqueeze(1).to(x.dtype)
        x = self.encoder(x, x_mask)
        stats = self.proj(x) * x_mask

        m, logs = torch.split(stats, self.out_channels, dim=1)
        return m, logs, x_mask


class PosteriorEncoder(torch.nn.Module):
    """
    Posterior Encoder for inferring latent representation.

    Args:
        in_channels (int): Number of channels in the input.
        out_channels (int): Number of channels in the output.
        hidden_channels (int): Number of hidden channels in the encoder.
        kernel_size (int): Kernel size of the convolutional layers.
        dilation_rate (int): Dilation rate of the convolutional layers.
        n_layers (int): Number of layers in the encoder.
        gin_channels (int, optional): Number of channels for the global conditioning input. Defaults to 0.

    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        hidden_channels: int,
        kernel_size: int,
        dilation_rate: int,
        n_layers: int,
        gin_channels: int = 0,
    ):
        super().__init__()
        self.out_channels = out_channels
        self.pre = torch.nn.Conv1d(in_channels, hidden_channels, 1)
        self.enc = WaveNet(
            hidden_channels,
            kernel_size,
            dilation_rate,
            n_layers,
            gin_channels=gin_channels,
        )
        self.proj = torch.nn.Conv1d(hidden_channels, out_channels * 2, 1)

    def forward(
        self,
        x: torch.Tensor,
        x_lengths: torch.Tensor,
        g: torch.Tensor | None = None,
    ):
        x_mask = sequence_mask(x_lengths, x.size(2)).unsqueeze(1).to(x.dtype)

        x = self.pre(x) * x_mask
        x = self.enc(x, x_mask, g=g)

        stats = self.proj(x) * x_mask
        m, logs = torch.split(stats, self.out_channels, dim=1)

        z = m + torch.randn_like(m) * torch.exp(logs)
        z *= x_mask

        return z, m, logs, x_mask

    def remove_weight_norm(self):
        self.enc.remove_weight_norm()

    def __prepare_scriptable__(self):
        for hook in self.enc._forward_pre_hooks.values():
            if (
                hook.__module__ == "torch.nn.utils.parametrizations.weight_norm"
                and hook.__class__.__name__ == "WeightNorm"
            ):
                torch.nn.utils.remove_weight_norm(self.enc)
        return self



"""
Module which defines the command-line interface for training voice
models using RVC.
"""

from __future__ import annotations

from typing import Annotated

import time
from multiprocessing import cpu_count
from pathlib import Path  # noqa: TC003

import typer
from rich import print as rprint
from rich.panel import Panel
from rich.table import Table

from ultimate_rvc.cli.common import (
    complete_audio_split_method,
    complete_device_type,
    complete_embedder_model,
    complete_f0_method,
    complete_index_algorithm,
    complete_pretrained_type,
    complete_training_sample_rate,
    complete_vocoder,
    format_duration,
)
from ultimate_rvc.cli.typing_extra import PanelName
from ultimate_rvc.core.train.common import get_gpu_info as _get_gpu_info
from ultimate_rvc.core.train.extract import extract_features as _extract_features
from ultimate_rvc.core.train.prepare import populate_dataset as _populate_dataset
from ultimate_rvc.core.train.prepare import preprocess_dataset as _preprocess_dataset
from ultimate_rvc.core.train.train import run_training as _run_training
from ultimate_rvc.typing_extra import (
    AudioSplitMethod,
    DeviceType,
    EmbedderModel,
    IndexAlgorithm,
    PretrainedType,
    TrainingF0Method,
    TrainingSampleRate,
    Vocoder,
)

app = typer.Typer(
    name="train",
    no_args_is_help=True,
    help="Train voice models using RVC",
    rich_markup_mode="markdown",
)

CORES = cpu_count()


@app.command(no_args_is_help=True)
def populate_dataset(
    name: Annotated[
        str,
        typer.Argument(help="The name of the dataset to populate."),
    ],
    audio_files: Annotated[
        list[Path],
        typer.Argument(
            help="The audio files to populate the dataset with.",
            exists=True,
            dir_okay=False,
            resolve_path=True,
            file_okay=True,
        ),
    ],
) -> None:
    """
    Populate the dataset with the provided name with the provided audio
    files.
    """
    start_time = time.perf_counter()

    dataset_path = _populate_dataset(name, audio_files)

    rprint("[+] Dataset succesfully populated with the provided audio files!")
    rprint()
    rprint("Elapsed time:", format_duration(time.perf_counter() - start_time))
    rprint(Panel(f"[green]{dataset_path}", title="Dataset Path"))


@app.command(no_args_is_help=True)
def preprocess_dataset(
    model_name: Annotated[
        str,
        typer.Argument(
            help=(
                "The name of the voice model to train. If no voice model with the"
                " provided name exists for training, a new voice model for training"
                " will be created with the provided name. If a voice model with the"
                " provided name already exists for training, then its currently"
                " associated dataset will be replaced with the provided dataset."
            ),
        ),
    ],
    dataset: Annotated[
        Path,
        typer.Argument(
            help="The path to the dataset to preprocess",
            exists=True,
            resolve_path=True,
            file_okay=False,
            dir_okay=True,
        ),
    ],
    sample_rate: Annotated[
        TrainingSampleRate,
        typer.Option(
            autocompletion=complete_training_sample_rate,
            help="The target sample rate for the audio files in the provided dataset",
        ),
    ] = TrainingSampleRate.HZ_40K,
    split_method: Annotated[
        AudioSplitMethod,
        typer.Option(
            case_sensitive=False,
            autocompletion=complete_audio_split_method,
            help=(
                "The method to use for splitting the audio files in the provided"
                " dataset. Use the `Skip` method to skip splitting if the audio files"
                " are already split. Use the `Simple` method if excessive silence has"
                " already been removed from the audio files. Use the `Automatic` method"
                " for automatic silence detection and splitting around it."
            ),
        ),
    ] = AudioSplitMethod.AUTOMATIC,
    chunk_len: Annotated[
        float,
        typer.Option(
            min=0.5,
            max=5.0,
            help="Length of split audio chunks when using the `Simple` split method.",
        ),
    ] = 3.0,
    overlap_len: Annotated[
        float,
        typer.Option(
            min=0.0,
            max=0.4,
            help=(
                "Length of overlap between split audio chunks when using the `Simple`"
                " split method."
            ),
        ),
    ] = 0.3,
    filter_audio: Annotated[
        bool,
        typer.Option(
            help=(
                "Whether to remove low-frequency sounds from the audio files in the"
                " provided dataset by applying a high-pass butterworth filter."
            ),
        ),
    ] = True,
    clean_audio: Annotated[
        bool,
        typer.Option(
            help=(
                "Whether to clean the audio files in the provided dataset using noise"
                " reduction algorithms."
            ),
        ),
    ] = False,
    clean_strength: Annotated[
        float,
        typer.Option(
            min=0.0,
            max=1.0,
            help=(
                "The intensity of the cleaning to apply to the audio files in the"
                " provided dataset."
            ),
        ),
    ] = 0.7,
    cpu_cores: Annotated[
        int,
        typer.Option(
            min=1,
            max=CORES,
            help="The number of CPU cores to use for preprocessing",
        ),
    ] = CORES,
) -> None:
    """
    Preprocess a dataset of audio files for training a voice
    model.
    """
    start_time = time.perf_counter()

    _preprocess_dataset(
        model_name=model_name,
        dataset=dataset,
        sample_rate=sample_rate,
        split_method=split_method,
        chunk_len=chunk_len,
        overlap_len=overlap_len,
        filter_audio=filter_audio,
        clean_audio=clean_audio,
        clean_strength=clean_strength,
        cpu_cores=cpu_cores,
    )

    rprint("[+] Dataset succesfully preprocessed!")
    rprint()
    rprint("Elapsed time:", format_duration(time.perf_counter() - start_time))


@app.command()
def get_gpu_information() -> None:
    """Retrieve information on locally available GPUs."""
    start_time = time.perf_counter()
    rprint("[+] Retrieving GPU Information...")
    gpu_infos = _get_gpu_info()

    rprint("[+] GPU Information successfully retrieved!")
    rprint()
    rprint("Elapsed time:", format_duration(time.perf_counter() - start_time))

    table = Table()
    table.add_column("Name", style="green")
    table.add_column("Index", style="green")

    for gpu_name, gpu_index in gpu_infos:
        table.add_row(gpu_name, str(gpu_index))

    rprint(table)


@app.command(no_args_is_help=True)
def extract_features(
    model_name: Annotated[
        str,
        typer.Argument(help="The name of the voice model to be trained."),
    ],
    f0_method: Annotated[
        TrainingF0Method,
        typer.Option(
            case_sensitive=False,
            autocompletion=complete_f0_method,
            help="The method to use for extracting pitch features.",
        ),
    ] = TrainingF0Method.RMVPE,
    hop_length: Annotated[
        int,
        typer.Option(
            min=1,
            max=512,
            help=(
                "The hop length to use for extracting pitch features. Only used"
                " with the CREPE pitch extraction method."
            ),
        ),
    ] = 128,
    embedder_model: Annotated[
        EmbedderModel,
        typer.Option(
            autocompletion=complete_embedder_model,
            help="The model to use for extracting audio embeddings.",
            case_sensitive=False,
        ),
    ] = EmbedderModel.CONTENTVEC,
    custom_embedder_model: Annotated[
        str | None,
        typer.Option(
            exists=True,
            resolve_path=True,
            dir_okay=True,
            file_okay=False,
            help="The path to a custom model to use for extracting audio embeddings.",
        ),
    ] = None,
    include_mutes: Annotated[
        int,
        typer.Option(
            help=(
                "The number of mute audio files to include in the generated"
                " training file list. Adding silent files enables the voice model to"
                " handle pure silence in inferred audio files. If the preprocessed"
                " audio dataset already contains segments of pure silence, set this"
                " to 0."
            ),
            min=0,
            max=10,
        ),
    ] = 2,
    cpu_cores: Annotated[
        int,
        typer.Option(
            help="The number of CPU cores to use for feature extraction.",
            min=1,
            max=cpu_count(),
        ),
    ] = cpu_count(),
    hardware_acceleration: Annotated[
        DeviceType,
        typer.Option(
            autocompletion=complete_device_type,
            case_sensitive=False,
            help=(
                "The type of hardware acceleration to use for feature extraction."
                " `AUTOMATIC` will automatically select the first available GPU and"
                " fall back to CPU if no GPUs are available."
            ),
        ),
    ] = DeviceType.AUTOMATIC,
    gpu_id: Annotated[
        list[int] | None,
        typer.Option(
            min=0,
            help=(
                "The id of a GPU to use for feature extraction when `GPU` is selected"
                " for hardware acceleration. This option can be provided multiple times"
                " to use multiple GPUs in parallel."
            ),
        ),
    ] = None,
) -> None:
    """
    Extract features from the preprocessed dataset associated with a
    voice model to be trained.
    """
    start_time = time.perf_counter()

    gpu_id_set = set(gpu_id) if gpu_id is not None else None
    _extract_features(
        model_name=model_name,
        f0_method=f0_method,
        hop_length=hop_length,
        embedder_model=embedder_model,
        custom_embedder_model=custom_embedder_model,
        include_mutes=include_mutes,
        cpu_cores=cpu_cores,
        hardware_acceleration=hardware_acceleration,
        gpu_ids=gpu_id_set,
    )

    rprint("[+] Dataset features succesfully extracted!")
    rprint()
    rprint("Elapsed time:", format_duration(time.perf_counter() - start_time))


@app.command(no_args_is_help=True)
def run_training(
    model_name: Annotated[
        str,
        typer.Argument(
            help="The name of the voice model to train.",
        ),
    ],
    num_epochs: Annotated[
        int,
        typer.Option(
            rich_help_panel=PanelName.TRAINING_OPTIONS,
            help=(
                "The number of epochs to train the voice model.A higher number can"
                " improve voice model performance but may lead to overtraining."
            ),
            min=1,
        ),
    ] = 500,
    batch_size: Annotated[
        int,
        typer.Option(
            rich_help_panel=PanelName.TRAINING_OPTIONS,
            help=(
                "The number of samples to include in each training batch. It is"
                " advisable to align this value with the available VRAM of your GPU. A"
                " setting of 4 offers improved accuracy but slower processing, while 8"
                " provides faster and standard results."
            ),
            min=1,
        ),
    ] = 8,
    detect_overtraining: Annotated[
        bool,
        typer.Option(
            rich_help_panel=PanelName.TRAINING_OPTIONS,
            help=(
                "Whether to detect overtraining to prevent the voice model from"
                " learning the training data too well and losing the ability to"
                " generalize to new data."
            ),
        ),
    ] = False,
    overtraining_threshold: Annotated[
        int,
        typer.Option(
            rich_help_panel=PanelName.TRAINING_OPTIONS,
            help=(
                "The maximum number of epochs to continue training without any observed"
                " improvement in model performance."
            ),
            min=1,
        ),
    ] = 50,
    vocoder: Annotated[
        Vocoder,
        typer.Option(
            rich_help_panel=PanelName.ALGORITHMIC_OPTIONS,
            case_sensitive=False,
            autocompletion=complete_vocoder,
            help=(
                "The vocoder to use for audio synthesis during training. HiFi-GAN"
                " provides basic audio fidelity, while RefineGAN provides the highest"
                " audio fidelity."
            ),
        ),
    ] = Vocoder.HIFI_GAN,
    index_algorithm: Annotated[
        IndexAlgorithm,
        typer.Option(
            rich_help_panel=PanelName.ALGORITHMIC_OPTIONS,
            case_sensitive=False,
            autocompletion=complete_index_algorithm,
            help=(
                "The method to use for generating an index file for the trained voice"
                " model. KMeans is particularly useful for large datasets."
            ),
        ),
    ] = IndexAlgorithm.AUTO,
    pretrained_type: Annotated[
        PretrainedType,
        typer.Option(
            rich_help_panel=PanelName.ALGORITHMIC_OPTIONS,
            case_sensitive=False,
            autocompletion=complete_pretrained_type,
            help=(
                "The type of pretrained model to finetune the voice model on."
                " `None` will train the voice model from scratch, while `Default` will"
                " use a pretrained model tailored to the specific voice model"
                " architecture. `Custom` will use a custom pretrained model that you"
                " provide."
            ),
        ),
    ] = PretrainedType.DEFAULT,
    custom_pretrained: Annotated[
        str | None,
        typer.Option(
            rich_help_panel=PanelName.ALGORITHMIC_OPTIONS,
            help=(
                "The name of a custom pretrained model to finetune the voice model on."
            ),
        ),
    ] = None,
    save_interval: Annotated[
        int,
        typer.Option(
            rich_help_panel=PanelName.DATA_STORAGE_OPTIONS,
            help=(
                "The epoch interval at which to to save voice model weights and"
                " checkpoints. The best model weights are always saved regardless of"
                " this setting."
            ),
            min=1,
        ),
    ] = 10,
    save_all_checkpoints: Annotated[
        bool,
        typer.Option(
            rich_help_panel=PanelName.DATA_STORAGE_OPTIONS,
            help=(
                " Whether to save a unique checkpoint at each save interval. If not"
                " enabled, only the latest checkpoint will be saved at each interval."
            ),
        ),
    ] = False,
    save_all_weights: Annotated[
        bool,
        typer.Option(
            rich_help_panel=PanelName.DATA_STORAGE_OPTIONS,
            help=(
                "Whether to save unique voice model weights at each save interval. If"
                " not enabled, only the best voice model weights will be saved."
            ),
        ),
    ] = False,
    clear_saved_data: Annotated[
        bool,
        typer.Option(
            rich_help_panel=PanelName.DATA_STORAGE_OPTIONS,
            help=(
                "Whether to delete any existing training data associated with the"
                " voice model before training commences. Enable this setting only if"
                " you are training a new voice model from scratch or restarting"
                " training."
            ),
        ),
    ] = False,
    upload_model: Annotated[
        bool,
        typer.Option(
            rich_help_panel=PanelName.DATA_STORAGE_OPTIONS,
            help=(
                "Whether to automatically upload the trained voice model so that it"
                " can be used for generation tasks audio generation tasks within the"
                " Ultimate RVC app."
            ),
        ),
    ] = False,
    upload_name: Annotated[
        str | None,
        typer.Option(
            rich_help_panel=PanelName.DATA_STORAGE_OPTIONS,
            help="The name to give the uploaded voice model.",
        ),
    ] = None,
    hardware_acceleration: Annotated[
        DeviceType,
        typer.Option(
            rich_help_panel=PanelName.DEVICE_AND_MEMORY_OPTIONS,
            autocompletion=complete_device_type,
            case_sensitive=False,
            help=(
                "The type of hardware acceleration to use for training the voice model."
                "`AUTOMATIC` will automatically select the first available GPU and fall"
                " back to CPU if no GPUs are available."
            ),
        ),
    ] = DeviceType.AUTOMATIC,
    gpu_id: Annotated[
        list[int] | None,
        typer.Option(
            rich_help_panel=PanelName.DEVICE_AND_MEMORY_OPTIONS,
            min=0,
            help=(
                "The id of a GPU to use for training the voice model when `GPU` is"
                " selected for hardware acceleration. This option can be provided"
                " multiple times to use multiple GPUs in parallel."
            ),
        ),
    ] = None,
    preload_dataset: Annotated[
        bool,
        typer.Option(
            rich_help_panel=PanelName.DEVICE_AND_MEMORY_OPTIONS,
            help=(
                "Whether to preload all training data into GPU memory. This can improve"
                " training speed but requires a lot of VRAM."
            ),
        ),
    ] = False,
    reduce_memory_usage: Annotated[
        bool,
        typer.Option(
            rich_help_panel=PanelName.DEVICE_AND_MEMORY_OPTIONS,
            help=(
                "Whether to reduce VRAM usage at the cost of slower training speed by"
                " enabling activation checkpointing. This is useful for GPUs with"
                " limited memory (e.g., <6GB VRAM) or when training with a batch size"
                " larger than what your GPU can normally accommodate."
            ),
        ),
    ] = False,
) -> None:
    """
    Train a voice model using its associated preprocessed dataset and
    extracted features.
    """
    start_time = time.perf_counter()

    gpu_id_set = set(gpu_id) if gpu_id is not None else None
    trained_model_files = _run_training(
        model_name=model_name,
        num_epochs=num_epochs,
        batch_size=batch_size,
        detect_overtraining=detect_overtraining,
        overtraining_threshold=overtraining_threshold,
        vocoder=vocoder,
        index_algorithm=index_algorithm,
        pretrained_type=pretrained_type,
        custom_pretrained=custom_pretrained,
        save_interval=save_interval,
        save_all_checkpoints=save_all_checkpoints,
        save_all_weights=save_all_weights,
        clear_saved_data=clear_saved_data,
        upload_model=upload_model,
        upload_name=upload_name,
        hardware_acceleration=hardware_acceleration,
        gpu_ids=gpu_id_set,
        preload_dataset=preload_dataset,
        reduce_memory_usage=reduce_memory_usage,
    )
    if trained_model_files is None:
        rprint("[!] Training failed!")
        return
    model_file, index_file = trained_model_files

    rprint("[+] Voice model succesfully trained!")
    rprint()
    rprint("Elapsed time:", format_duration(time.perf_counter() - start_time))
    rprint(Panel(f"[green]{model_file}", title="Model File"))
    rprint(Panel(f"[green]{index_file}", title="Index File"))



"""The Ultimate RVC project."""

from __future__ import annotations

import logging
import os
from logging.handlers import RotatingFileHandler
from pathlib import Path

from ultimate_rvc.common import BASE_DIR

logger = logging.getLogger()

URVC_NO_LOGGING = os.getenv("URVC_NO_LOGGING", "0") == "1"
URVC_LOGS_DIR = Path(os.getenv("URVC_LOGS_DIR") or BASE_DIR / "logs")
URVC_CONSOLE_LOG_LEVEL = os.getenv("URVC_CONSOLE_LOG_LEVEL", "ERROR")
URVC_FILE_LOG_LEVEL = os.getenv("URVC_FILE_LOG_LEVEL", "INFO")

if URVC_NO_LOGGING:
    logging.basicConfig(handlers=[logging.NullHandler()])

else:
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(URVC_CONSOLE_LOG_LEVEL)

    URVC_LOGS_DIR.mkdir(exist_ok=True, parents=True)
    file_handler = RotatingFileHandler(
        URVC_LOGS_DIR / "ultimate_rvc.log",
        mode="a",
        maxBytes=1024 * 1024 * 5,
        backupCount=1,
        encoding="utf-8",
    )
    file_handler.setLevel(URVC_FILE_LOG_LEVEL)

    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        style="%",
        level=logging.DEBUG,
        handlers=[stream_handler, file_handler],
    )



"""
Module which defines functions for initializing the core of the Ultimate
RVC project.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import lazy_loader as lazy

from pathlib import Path

from rich import print as rprint

from ultimate_rvc.common import VOICE_MODELS_DIR
from ultimate_rvc.core.common import FLAG_FILE
from ultimate_rvc.core.generate.song_cover import initialize_audio_separator
from ultimate_rvc.core.manage.models import download_voice_model
from ultimate_rvc.rvc.lib.tools.prerequisites_download import (
    prequisites_download_pipeline,
)

if TYPE_CHECKING:
    import static_sox

else:
    static_sox = lazy.load("static_sox")


def download_sample_models() -> None:
    """Download sample RVC models."""
    named_model_links = [
        (
            "https://huggingface.co/damnedraxx/TaylorSwift/resolve/main/TaylorSwift.zip",
            "Taylor Swift",
        ),
        (
            "https://huggingface.co/Vermiculos/balladjames/resolve/main/Ballad%20James.zip?download=true",
            "James Hetfield",
        ),
        ("https://huggingface.co/ryolez/MMLP/resolve/main/MMLP.zip", "Eminem"),
    ]
    for model_url, model_name in named_model_links:
        if not Path(VOICE_MODELS_DIR / model_name).is_dir():
            rprint(f"Downloading {model_name}...")
            try:
                download_voice_model(model_url, model_name)
            except Exception as e:  # noqa: BLE001
                rprint(f"Failed to download {model_name}: {e}")


def initialize() -> None:
    """Initialize the Ultimate RVC project."""
    prequisites_download_pipeline(exe=False)
    if not FLAG_FILE.is_file():
        # NOTE we only add_paths so that sox
        # binaries are downloaded as part of initialization.
        static_sox.add_paths(weak=True)
        download_sample_models()
        initialize_audio_separator()
        FLAG_FILE.touch()


if __name__ == "__main__":
    initialize()



import torch


class LayerNorm(torch.nn.Module):
    """
    Layer normalization module.

    Args:
        channels (int): Number of channels.
        eps (float, optional): Epsilon value for numerical stability. Defaults to 1e-5.

    """

    def __init__(self, channels: int, eps: float = 1e-5):
        super().__init__()
        self.eps = eps
        self.gamma = torch.nn.Parameter(torch.ones(channels))
        self.beta = torch.nn.Parameter(torch.zeros(channels))

    def forward(self, x):
        # Transpose to (batch_size, time_steps, channels) for layer_norm
        x = x.transpose(1, -1)
        x = torch.nn.functional.layer_norm(
            x,
            (x.size(-1),),
            self.gamma,
            self.beta,
            self.eps,
        )
        # Transpose back to (batch_size, channels, time_steps)
        return x.transpose(1, -1)



from typing import Literal, Self, overload

import numpy as np
from numpy.typing import NDArray

class AudioFile:
    @classmethod
    @overload
    def __new__(
        cls: object,
        filename: str,
        mode: Literal["r"] = "r",
    ) -> ReadableAudioFile: ...
    @classmethod
    @overload
    def __new__(
        cls: object,
        filename: str,
        mode: Literal["w"],
        samplerate: float | None = None,
        num_channels: int = 1,
        bit_depth: int = 16,
        quality: str | float | None = None,
    ) -> WriteableAudioFile: ...

class ReadableAudioFile(AudioFile):
    def __enter__(self) -> Self: ...
    def __exit__(self, arg0: object, arg1: object, arg2: object) -> None: ...
    def read(self, num_frames: float = 0) -> NDArray[np.float32]: ...
    def tell(self) -> int: ...
    @property
    def frames(self) -> int: ...
    @property
    def num_channels(self) -> int: ...
    @property
    def samplerate(self) -> float | int: ...

class WriteableAudioFile(AudioFile):
    def __enter__(self) -> Self: ...
    def __exit__(self, arg0: object, arg1: object, arg2: object) -> None: ...
    def write(self, samples: NDArray[...]) -> None: ...



import concurrent.futures
import glob
import json
import logging
import multiprocessing as mp
import os
import sys
import time

import numpy as np
import tqdm

import torch
import torchcrepe

now_dir = os.getcwd()
sys.path.append(os.path.join(now_dir))

# Zluda hijack
import ultimate_rvc.rvc.lib.zluda
from ultimate_rvc.common import RVC_MODELS_DIR
from ultimate_rvc.rvc.configs.config import Config
from ultimate_rvc.rvc.lib.predictors.RMVPE import RMVPE0Predictor
from ultimate_rvc.rvc.lib.utils import load_audio, load_embedding
from ultimate_rvc.rvc.train.utils import remove_sox_libmso6_from_ld_preload

logger = logging.getLogger(__name__)

# Load config
config = Config()
mp.set_start_method("spawn", force=True)


class FeatureInput:

    def __init__(self, sample_rate=16000, hop_size=160, device="cpu"):
        self.fs = sample_rate
        self.hop = hop_size
        self.f0_bin = 256
        self.f0_max = 1100.0
        self.f0_min = 50.0
        self.f0_mel_min = 1127 * np.log(1 + self.f0_min / 700)
        self.f0_mel_max = 1127 * np.log(1 + self.f0_max / 700)
        self.device = device
        self.model_rmvpe = None

    def compute_f0(self, audio_array, method, hop_length):
        if method == "crepe":
            return self._get_crepe(audio_array, hop_length, type="full")
        if method == "crepe-tiny":
            return self._get_crepe(audio_array, hop_length, type="tiny")
        if method == "rmvpe":
            return self.model_rmvpe.infer_from_audio(audio_array, thred=0.03)
        raise ValueError(f"Unknown F0 method: {method}")

    def _get_crepe(self, x, hop_length, type):
        audio = torch.from_numpy(x.astype(np.float32)).to(self.device)
        audio /= torch.quantile(torch.abs(audio), 0.999)
        audio = audio.unsqueeze(0)
        pitch = torchcrepe.predict(
            audio,
            self.fs,
            hop_length,
            self.f0_min,
            self.f0_max,
            type,
            batch_size=hop_length * 2,
            device=audio.device,
            pad=True,
        )
        source = pitch.squeeze(0).cpu().float().numpy()
        source[source < 0.001] = np.nan
        return np.nan_to_num(
            np.interp(
                np.arange(0, len(source) * (x.size // self.hop), len(source))
                / (x.size // self.hop),
                np.arange(0, len(source)),
                source,
            ),
        )

    def coarse_f0(self, f0):
        f0_mel = 1127.0 * np.log(1.0 + f0 / 700.0)
        f0_mel = np.clip(
            (f0_mel - self.f0_mel_min)
            * (self.f0_bin - 2)
            / (self.f0_mel_max - self.f0_mel_min)
            + 1,
            1,
            self.f0_bin - 1,
        )
        return np.rint(f0_mel).astype(int)

    def process_file(self, file_info, f0_method, hop_length):
        inp_path, opt_path_coarse, opt_path_full, _ = file_info

        if os.path.exists(opt_path_coarse) and os.path.exists(opt_path_full):
            return

        try:
            np_arr = load_audio(inp_path, self.fs)
            feature_pit = self.compute_f0(np_arr, f0_method, hop_length)
            np.save(opt_path_full, feature_pit, allow_pickle=False)
            coarse_pit = self.coarse_f0(feature_pit)
            np.save(opt_path_coarse, coarse_pit, allow_pickle=False)
        except Exception as error:
            logger.error(  # noqa: TRY400
                "An error occurred extracting file %s on %s: %s",
                inp_path,
                self.device,
                error,
            )

    def process_files(self, files, f0_method, hop_length, device, threads):
        self.device = device
        if f0_method == "rmvpe":
            self.model_rmvpe = RMVPE0Predictor(
                os.path.join(str(RVC_MODELS_DIR), "predictors", "rmvpe.pt"),
                device=device,
            )

        def worker(file_info):
            self.process_file(file_info, f0_method, hop_length)

        with tqdm.tqdm(total=len(files), leave=True) as pbar:
            with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
                futures = [executor.submit(worker, f) for f in files]
                for _ in concurrent.futures.as_completed(futures):
                    pbar.update(1)


def run_pitch_extraction(
    files: list[list[str]],
    devices: list[str],
    f0_method: str,
    hop_length: int,
    threads: int,
) -> None:
    devices_str = ", ".join(devices)
    logger.info(
        "Starting pitch extraction with %d cores on %s using %s...",
        threads,
        devices_str,
        f0_method,
    )
    start_time = time.time()
    fe = FeatureInput()

    remove_sox_libmso6_from_ld_preload()

    if f0_method == "crepe":
        actual_threads = 1  # crepe method cannot handle multiple threads
        logger.info(
            "crepe detected: using single-threaded processing (was %d threads). "
            "crepe requires single-threaded operation due to GPU limitations.",
            threads,
        )
    else:
        actual_threads = threads
    with concurrent.futures.ProcessPoolExecutor(max_workers=len(devices)) as executor:
        tasks = [
            executor.submit(
                fe.process_files,
                files[i :: len(devices)],
                f0_method,
                hop_length,
                devices[i],
                actual_threads // len(devices),
            )
            for i in range(len(devices))
        ]
        for future in concurrent.futures.as_completed(tasks):
            future.result()  # Properly waits and propagates exceptions

    logger.info("Pitch extraction completed in %.2f seconds.", time.time() - start_time)


def process_file_embedding(
    files,
    embedder_model,
    embedder_model_custom,
    device_num,
    device,
    n_threads,
):
    model = load_embedding(embedder_model, embedder_model_custom).to(device).float()
    model.eval()
    n_threads = max(1, n_threads)

    def worker(file_info):
        wav_file_path, _, _, out_file_path = file_info
        if os.path.exists(out_file_path):
            return
        feats = torch.from_numpy(load_audio(wav_file_path, 16000)).to(device).float()
        feats = feats.view(1, -1)
        with torch.no_grad():
            result = model(feats)["last_hidden_state"]
        feats_out = result.squeeze(0).float().cpu().numpy()
        if not np.isnan(feats_out).any():
            np.save(out_file_path, feats_out, allow_pickle=False)
        else:
            logger.error("%s contains NaN values and will be skipped.", wav_file_path)

    with tqdm.tqdm(total=len(files), leave=True, position=device_num) as pbar:
        with concurrent.futures.ThreadPoolExecutor(max_workers=n_threads) as executor:
            futures = [executor.submit(worker, f) for f in files]
            for _ in concurrent.futures.as_completed(futures):
                pbar.update(1)


def run_embedding_extraction(
    files: list[list[str]],
    devices: list[str],
    embedder_model: str,
    embedder_model_custom: str | None,
    threads: int,
) -> None:
    devices_str = ", ".join(devices)
    logger.info(
        "Starting embedding extraction with %d cores on %s...",
        threads,
        devices_str,
    )
    start_time = time.time()
    with concurrent.futures.ProcessPoolExecutor(max_workers=len(devices)) as executor:
        tasks = [
            executor.submit(
                process_file_embedding,
                files[i :: len(devices)],
                embedder_model,
                embedder_model_custom,
                i,
                devices[i],
                threads // len(devices),
            )
            for i in range(len(devices))
        ]
        for future in concurrent.futures.as_completed(tasks):
            future.result()  # Properly waits and propagates exceptions
    logger.info(
        "Embedding extraction completed in %.2f seconds.",
        time.time() - start_time,
    )


def initialize_extraction(
    exp_dir: str,
    f0_method: str,
    embedder_model: str,
) -> list[list[str]]:
    wav_path = os.path.join(exp_dir, "sliced_audios_16k")
    os.makedirs(os.path.join(exp_dir, f"f0_{f0_method}"), exist_ok=True)
    os.makedirs(os.path.join(exp_dir, f"f0_{f0_method}_voiced"), exist_ok=True)
    os.makedirs(
        os.path.join(exp_dir, f"{embedder_model}_extracted"),
        exist_ok=True,
    )

    files: list[list[str]] = []
    for file in glob.glob(os.path.join(wav_path, "*.wav")):
        file_name = os.path.basename(file)
        file_info = [
            file,
            os.path.join(exp_dir, f"f0_{f0_method}", file_name + ".npy"),
            os.path.join(exp_dir, f"f0_{f0_method}_voiced", file_name + ".npy"),
            os.path.join(
                exp_dir,
                f"{embedder_model}_extracted",
                file_name.replace("wav", "npy"),
            ),
        ]
        files.append(file_info)

    return files


def update_model_info(
    exp_dir: str,
    embedder_model: str,
    custom_embedder_model_hash: str | None,
) -> None:
    file_path = os.path.join(exp_dir, "model_info.json")
    if os.path.exists(file_path):
        with open(file_path) as f:
            data = json.load(f)
    else:
        data = {}
    data["embedder_model"] = embedder_model
    data["custom_embedder_model_hash"] = custom_embedder_model_hash
    with open(file_path, "w") as f:
        json.dump(data, f, indent=4)
